# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

import sys
from typing import Any, TextIO, Union

if sys.version_info < (3, 8):
    from typing_extensions import Protocol, runtime_checkable
else:
    from typing import Protocol, runtime_checkable

if sys.version_info < (3, 9):
    from importlib_resources.abc import Traversable
else:
    from importlib.abc import Traversable


__all__ = (
    "Protocol",
    "runtime_checkable",
    "Traversable",
    "StringOrIO",
    "HasOpen",
    "HasRead",
)


StringOrIO = Union[Traversable, str, TextIO]


@runtime_checkable
class HasOpen(Protocol):
    def open(self) -> Any:
        pass


@runtime_checkable
class HasRead(Protocol):
    def read(self) -> str:
        pass



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

import contextlib
import csv
import sys
from collections.abc import Mapping
from typing import Any, Callable, Generic, Iterator, TextIO, TypeVar, Union, overload

from .. import data
from ..exceptions import MatchingIDNotFound
from ..typing import HasOpen, HasRead, StringOrIO

A = TypeVar("A")
B = TypeVar("B")
A_conv = Callable[[str], Union[A, int]]
B_conv = Callable[[str], Union[B, int]]


class BiMap(Generic[A, B]):
    def __init__(
        self,
        class_A: type[A],
        class_B: type[B],
        converters: tuple[A_conv, B_conv] = (int, int),  # type: ignore[type-arg]
        filename: StringOrIO | None = None,
    ) -> None:
        """
        Bi-bidirectional map class.

        Parameters
        ----------
        class_A, class_B: class types
            Input class types.
        converters: tuple, optional, default=(int,int)
            Converter functions applied on each entry (row) of the file
            providing the class_a-class_B matches.
            The order of the list elements must agree with that of the classes
            passed in in the constructor.
            By default, data on the file is assumed to be integers,
            which is the typical case for PDG IDs and alike.
        filename: string or file object, optional,
                  default='<class_B_name>_to_<class_A_name>.csv',
                  where the names are taken as lowercase.
            Specify a file from which to read all class_a-class_B matches.
            It does not matter whether the file contains the input as
            val_A,val_B or val_B,val_A.

        Examples
        --------
        >>> from particle import PDGID, PythiaID

        Basic/standard usage:
        >>> bimap = BiMap(PDGID, PythiaID)

        >>> bimap[PDGID(9010221)]
        <PythiaID: 10221>
        >>> bimap[PythiaID(10221)]
        <PDGID: 9010221>

        Advanced usage:
        >>> # Either pass a file name or a file object
        >>> from particle import data
        >>> filename = data.basepath / "pdgid_to_pythiaid.csv"
        >>> bimap = BiMap(PDGID, PythiaID, filename=filename)
        """

        self.class_A: type[A] = class_A
        self.class_B: type[B] = class_B

        name_A = self.class_A.__name__.upper()
        name_B = self.class_B.__name__.upper()

        file_object: TextIO
        if filename is None:
            filename = f"{name_A.lower()}_to_{name_B.lower()}.csv"
            file_object = data.basepath.joinpath(filename).open()
        elif isinstance(filename, HasRead):
            file_object = filename
        elif isinstance(filename, HasOpen):
            file_object = filename.open()
        else:
            file_object = open(filename, encoding="utf_8")  # type: ignore[arg-type]

        with file_object as _f:
            self._to_map = {
                converters[1](v[name_B]): converters[0](v[name_A])
                for v in csv.DictReader(line for line in _f if not line.startswith("#"))
            }
            _f.seek(0)
            self._from_map = {
                converters[0](v[name_A]): converters[1](v[name_B])
                for v in csv.DictReader(line for line in _f if not line.startswith("#"))
            }

    @overload
    def __getitem__(self, value: A) -> B:
        pass

    @overload
    def __getitem__(self, value: B) -> A:
        pass

    def __getitem__(self, value: Any) -> Any:
        if isinstance(value, self.class_B):
            with contextlib.suppress(KeyError):
                return self.class_A(self._to_map[value])  # type: ignore[call-arg]
        elif isinstance(value, self.class_A):
            with contextlib.suppress(KeyError):
                return self.class_B(self._from_map[value])  # type: ignore[call-arg]

        name_A = self.class_A.__name__
        name_B = self.class_B.__name__
        msg = f"Matching {name_A}-{name_B} for input {value} not found !"
        raise MatchingIDNotFound(msg)

    def __repr__(self) -> str:
        name_A = self.class_A.__name__
        name_B = self.class_B.__name__

        return f"<{self.__class__.__name__}({name_A}-{name_B}): {len(self)} matches>"

    def __len__(self) -> int:
        """Returns the number of matches."""
        return len(self._to_map)


def DirectionalMaps(
    name_A: str,
    name_B: str,
    converters: tuple[Callable[[str], str], Callable[[str], str]] = (str, str),
    filename: StringOrIO | None = None,
) -> tuple[DirectionalMap, DirectionalMap]:
    """
    Directional map class providing a to and from mapping.

    Parameters
    ----------
    name_A, name_B: str
        Input names of information to be mapped.
    converters: tuple, optional, default=(str,str)
        Converter functions applied on each entry (row) of the file
        providing the name_a-name_B matches.
        The order of the list elements must agree with that of the
        object names passed in in the constructor.
        By default, data on the file is assumed to be strings,
        which is the typical case for particle names.
    filename: string or file object, optional, default='particle/data/conversions.csv'.
        Specify a file from which to read all name_a-name_B matches.
        It is assumed that the order of items in the file matches the order
        of arguments specified in the class constructor, hence val_A,val_B.

    Examples
    --------

    >>> from particle import data  # doctest: +SKIP
    >>> filename = data.basepath / "a_to_b.csv"  # doctest: +SKIP
    >>> A2BMap, B2AMap = DirectionalMaps('A', 'B', filename=filename)  # doctest: +SKIP
    """

    name_A = name_A.upper()
    name_B = name_B.upper()

    fieldnames = None
    file_object: TextIO
    if filename is None:
        file_object = data.basepath.joinpath("conversions.csv").open()
    elif isinstance(filename, HasOpen):
        file_object = filename.open()
    elif isinstance(filename, HasRead):
        file_object = filename
    else:
        file_object = open(filename, encoding="utf_8")  # type: ignore[arg-type]

    with file_object as _f:
        skipinitialspace = True

        to_map = {
            converters[1](v[name_B]): converters[0](v[name_A])
            for v in csv.DictReader(
                (line for line in _f if not line.startswith("#")),
                fieldnames=fieldnames,
                skipinitialspace=skipinitialspace,
            )
        }
        _f.seek(0)
        from_map = {
            converters[0](v[name_A]): converters[1](v[name_B])
            for v in csv.DictReader(
                (line for line in _f if not line.startswith("#")),
                fieldnames=fieldnames,
                skipinitialspace=skipinitialspace,
            )
        }

    return (
        DirectionalMap(name_A, name_B, from_map),
        DirectionalMap(name_B, name_A, to_map),
    )


if sys.version_info < (3, 9):
    StrStrMapping = Mapping
else:
    StrStrMapping = Mapping[str, str]


class DirectionalMap(StrStrMapping):
    # pylint: disable-next=redefined-builtin
    def __init__(self, name_A: str, name_B: str, map: dict[str, str]) -> None:
        """
        Directional map class providing a A -> B mapping.

        Parameters
        ----------
        name_A, name_B: str
            Input names of information to be mapped.
        map: dict
            Input mapping as a dictionary.
        """

        self.name_A = name_A.upper()
        self.name_B = name_B.upper()

        self._map = map

    def __getitem__(self, value: str) -> str:
        try:
            return self._map[value]
        except KeyError:
            msg = f"Matching {self.name_A}->{self.name_B} for input {value} not found !"
            raise MatchingIDNotFound(msg) from None

    def __iter__(self) -> Iterator[str]:
        return iter(self._map)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self.name_A}->{self.name_B}): {len(self)} matches>"

    def __len__(self) -> int:
        """Returns the number of matches."""
        return len(self._map)



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

from enum import IntEnum

import pytest


class PDGIDsEnum(IntEnum):
    """Sample of PDGIDs on which to run tests."""

    # Gauge and Higgs bosons
    Gluon = 21
    Photon = 22
    Z0 = 23
    WMinus = -24
    HiggsBoson = 25
    ZPrime = 32
    # Charged leptons
    Electron = 11
    Positron = -Electron
    Muon = 13
    AntiMuon = -Muon
    Tau = 15
    # Neutrinos
    Nu_e = 12
    NuBar_tau = -16
    # Quarks
    DQuark = 1
    UQuark = 2
    SQuark = 3
    CQuark = 4
    BQuark = 5
    TQuark = 6
    # 4th generation quarks and leptons
    TauPrime = 17
    BPrimeQuark = 7
    TPrimeQuark = 8
    # Quarkonia
    jpsi = 443
    psi_2S = 100443
    Upsilon_1S = 553
    Upsilon_4S = 300553
    # Light hadrons
    Pi0 = 111
    PiPlus = 211
    eta = 221
    eta_prime = 331
    a_0_1450_plus = 10211
    KL = 130
    KS = 310
    KMinus = -321
    rho_770_minus = -213
    phi = 333
    omega = 223
    K1_1270_0 = 10313
    K1_1400_0 = 20313
    rho_1700_0 = 30113
    a2_1320_minus = -215
    omega_3_1670 = 227
    f_4_2300 = 9010229  # example of a not-well-known meson
    Proton = 2212
    AntiNeutron = -2112
    Lambda = 3122
    Sigma0 = 3212
    SigmaPlus = 3222
    SigmaMinus = 3112
    Xi0 = 3322
    AntiXiMinus = -3312
    OmegaMinus = 3334
    # Charm hadrons
    D0 = 421
    DPlus = 411
    DsPlus = 431
    LcPlus = 4122
    # Beauty hadrons
    B0 = 511
    BPlus = 521
    Bs = 531
    BcPlus = 541
    Lb = 5122
    # Top hadrons
    T0 = 621
    LtPlus = 6122
    # Special particles
    Graviton = 39
    Reggeon = 110
    Pomeron = 990
    Odderon = 9990
    # Supersymmetric particles
    Gluino = 1000021
    Gravitino = 1000039
    STildeL = 1000003
    CTildeR = 2000004
    # R-hadrons
    RPlus_TTildeDbar = 1000612
    R0_GTildeG = 1000993
    RPlusPlus_GTildeUUU = 1092224
    # Q-balls
    QBall1 = 10000150
    QBall2 = -10000200
    # Dyons
    DyonSameMagElecChargeSign = 4110010
    DyonOppositeMagElecChargeSign = 4120010
    # Di-quarks
    DD1 = 1103
    SD0 = 3101
    # Nuclei
    HydrogenNucleus = 1000010010
    Carbon12 = 1000060120
    # Pentaquarks
    AntiUCbarCUDPentaquark = -9422144
    # example of spin 3/2 u-cbar-c-u-d pentaquark decaying to J/psi proton
    UCbarCUDPentaquark = 9422144
    # Technicolor
    Pi0TC = 3000111
    PiMinusTC = -3000211
    # Excited (composite) quarks and leptons
    UQuarkStar = 4000002
    AntiElectronStar = -4000011
    # Generator specific pseudoparticles or concepts
    AntiCHadron = -84
    # Invalid ID
    Invalid1 = 0  # illegal ID
    Invalid2 = 99999999  # general form is a 7-digit number


@pytest.fixture(scope="session")
def PDGIDs():
    return PDGIDsEnum



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

from particle.particle.enums import Charge, SpinType


def test_enums_Charge():
    assert Charge.p + Charge.m == Charge.o
    assert Charge.pp + Charge.mm == Charge.o


def test_enums_SpinType():
    assert SpinType.PseudoScalar == -SpinType.Scalar
    assert SpinType.Axial == -SpinType.Vector
    assert SpinType.PseudoTensor == -SpinType.Tensor



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

import pytest

from particle.exceptions import MatchingIDNotFound
from particle.pdgid import PDGID
from particle.pythia import PythiaID


def test_class_string_representations():
    pid = PythiaID(211)
    assert pid == 211
    assert pid.__str__() == "<PythiaID: 211>"


def test_class_return_type():
    assert isinstance(-PythiaID(211), PythiaID)
    assert isinstance(~PythiaID(211), PythiaID)


def test_class_inversion():
    assert -PythiaID(311) == ~PythiaID(311)


def test_from_pdgid():
    assert PythiaID.from_pdgid(9010221) == 10221

    assert PythiaID.from_pdgid(PDGID(9010221)) == 10221
    assert PythiaID.from_pdgid(PDGID(9010221)) == PythiaID(10221)

    with pytest.raises(MatchingIDNotFound):
        PythiaID.from_pdgid(9000221)


def test_to_pdgid():
    pythiaid = PythiaID(10331)
    assert pythiaid.to_pdgid() == 10221
    assert pythiaid.to_pdgid() == PDGID(10221)



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

from particle import PDGID, Corsika7ID, Particle
from particle.converters import Corsika72PDGIDBiMap


def test_Corsika72PDGID():
    pdgid = Corsika72PDGIDBiMap[Corsika7ID(5)]
    assert pdgid == -13

    cid = Corsika72PDGIDBiMap[PDGID(13)]
    assert cid.is_particle()
    assert cid == 6

    p = Particle.from_pdgid(cid.to_pdgid())
    # should be muon
    assert p.charge == -1



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

import pytest
from hepunits import meter, second
from pytest import approx

from particle import data
from particle.particle import Particle
from particle.particle.enums import Charge, Inv, Parity, SpinType, Status
from particle.particle.particle import InvalidParticle, ParticleNotFound
from particle.pdgid import PDGID
from particle.pdgid.functions import Location, _digit


def test_sorted_find():
    assert Particle.findall() == sorted(Particle.finditer())


def test_lambda_style_search():
    particles = Particle.findall(lambda p: p.pdg_name == "p")
    assert len(particles) == 4
    assert 2212 in particles
    assert -2212 in particles
    assert 1000010010 in particles
    assert -1000010010 in particles

    assert [
        p.pdgid for p in Particle.findall(lambda p: p.pdg_name == "p" and p > 0)
    ] == [
        2212,
        1000010010,
    ]
    assert [
        p.pdgid for p in Particle.findall(lambda p: p.pdg_name == "p" and p < 0)
    ] == [
        -2212,
        -1000010010,
    ]


def test_fuzzy_name_search():
    particles = Particle.findall("p~")
    assert len(particles) == 2
    assert -2212 in particles
    assert -1000010010 in particles


def test_keyword_style_search():
    particles = Particle.findall(pdg_name="p")
    assert len(particles) == 4
    assert 2212 in particles
    assert -2212 in particles
    assert 1000010010 in particles
    assert -1000010010 in particles


def test_keyword_style_search_with_except_catch():
    particles = Particle.findall(ctau=float("inf"))
    assert 11 in particles

    particles = Particle.findall(name="p")
    assert len(particles) == 2
    assert 2212 in particles
    assert 1000010010 in particles

    assert [p.pdgid for p in Particle.findall(pdg_name="p", particle=True)] == [
        2212,
        1000010010,
    ]
    assert [p.pdgid for p in Particle.findall(pdg_name="p", particle=False)] == [
        -2212,
        -1000010010,
    ]

    assert [p.pdgid for p in Particle.findall(name="p", particle=True)] == [
        2212,
        1000010010,
    ]
    assert [p.pdgid for p in Particle.findall(name="p~", particle=False)] == [
        -2212,
        -1000010010,
    ]


def test_keyword_lambda_style_search():
    particles = Particle.findall(pdg_name=lambda x: "p" == x)
    assert len(particles) == 4
    assert 2212 in particles
    assert -2212 in particles
    assert 1000010010 in particles
    assert -1000010010 in particles

    # Fuzzy name
    particles = Particle.findall(name=lambda x: "p" in x)
    assert len(particles) > 2
    assert 2212 in particles
    assert -2212 in particles

    # Name and particle
    assert len(Particle.findall(name=lambda x: x == "p", particle=True)) == 2

    # Unit based comparison
    assert 2212 in Particle.findall(lifetime=lambda x: x > 1 * second)


def test_complex_search():
    # Find all strange mesons with c*tau > 1 meter
    particles = Particle.findall(
        lambda p: p.pdgid.is_meson
        and p.pdgid.has_strange
        and p.width > 0
        and p.ctau > 1000.0,
        particle=True,
    )
    assert len(particles) == 2  # K+ and KL0
    assert 130 in particles
    assert 321 in particles

    # Find all strange anti-mesons with c*tau > 1 meter
    particles = Particle.findall(
        lambda p: p.pdgid.is_meson
        and p.pdgid.has_strange
        and p.width > 0
        and p.ctau > 1000.0,
        particle=False,
    )
    assert len(particles) == 1  # only the K-
    assert -321 in particles


def test_pdg():
    assert Particle.from_pdgid(211).pdgid == 211
    with pytest.raises(InvalidParticle):
        Particle.from_pdgid(0)


def test_pdg_convert():
    p = Particle.from_pdgid(211)
    assert isinstance(p.pdgid, PDGID)
    assert p.pdgid == 211
    assert int(p.pdgid) == 211


def test_nucleus_convert():
    p = Particle.from_nucleus_info(1, 2)
    assert p.pdgid == 1000010020
    p = Particle.from_nucleus_info(92, 235)
    assert p.pdgid == 1000922350
    p = Particle.from_nucleus_info(1, 2, anti=True)
    assert p.pdgid == -1000010020
    # No exited nuclei in database
    try:
        p = Particle.from_nucleus_info(1, 2, i=1)
        assert p.pdgid == 1000010021
    except ParticleNotFound:
        pass
    # No strange nuclei in database and strange PDGID not implemented
    try:
        p = Particle.from_nucleus_info(1, 2, l_strange=1)
        assert p.pdgid == 1100010020
    except ParticleNotFound and InvalidParticle:
        pass


def test_sorting():
    assert Particle.from_pdgid(211) < Particle.from_pdgid(311)
    assert Particle.from_pdgid(211) < Particle.from_pdgid(-311)


def test_int_compare():
    assert Particle.from_pdgid(211) > 0
    assert Particle.from_pdgid(-211) < 0
    assert Particle.from_pdgid(211) >= 0
    assert Particle.from_pdgid(-211) <= 0

    assert 0 < Particle.from_pdgid(211)
    assert 0 > Particle.from_pdgid(-211)
    assert 0 <= Particle.from_pdgid(211)
    assert 0 >= Particle.from_pdgid(-211)


def test_string():
    pi = Particle.from_string("pi+")
    assert pi.pdgid == 211

    with pytest.raises(ParticleNotFound):
        Particle.from_string("unknown")


def test_fuzzy_string():
    """
    The input name is not specific enough, in which case the search is done
    by pdg_name after failing a match by name.
    """
    p = Particle.from_string("a(0)(980)")  # all 3 charge stages match
    assert p.pdgid == 9000111


def test_str():
    pi = Particle.from_pdgid(211)
    assert str(pi) == "pi+"


def test_rep():
    pi = Particle.from_pdgid(211)
    assert "pdgid=211" in repr(pi)
    assert 'name="pi+"' in repr(pi)
    assert "mass=139.57" in repr(pi)


def test_basic_props():
    pi = Particle.from_pdgid(211)
    assert pi.pdg_name == "pi"
    assert pi.pdgid == 211
    assert pi.three_charge == Charge.p
    assert pi.charge == 1


def test_lifetime_props():
    pi = Particle.from_pdgid(211)
    assert pi.lifetime == approx(26.0327460625985)  # in nanoseconds
    assert pi.ctau == approx(7804.4209306)  # in millimeters


def test_charge_consistency():
    """
    The charge of a particle is presently stored in the CSV files
    (see Particle.charge for the motivation), but it can also be retrieved
    from the particle's PDG ID, *if* the latter is valid.
    This test makes sure both numbers are consistent for all particles in the PDG table.
    """
    for p in Particle.all():
        assert p.three_charge == p.pdgid.three_charge


def test_P_consistency_mesons():
    """
    The parity quantum number is stored in the (curated) data CSV files.
    For unflavoured mesons it can be calculated as P = (-1)^(L+1),
    and this relation can be checked against the CSV data.

    Note: mesons with PDGIDs of the kind 9XXXXXX (N=9) are not experimentally
    well-known particles and P is undefined.
    """
    for p in Particle.all():
        if not p.is_unflavoured_meson:
            continue
        elif _digit(p.pdgid, Location.N) == 9:
            continue
        elif p.pdgid == 22:  # Special case of the photon
            assert p.P == -1
        else:
            assert p.P == (-1) ** (p.L + 1)


def test_P_consistency_baryons():
    """
    The parity quantum number is stored in the (curated) data CSV files.
    For baryons the (intrinsic) parity flips sign for the antiparticle.
    As for baryons with undefined parity, that of the antibaryon
    is equally undefined, of course.
    """
    pdgid = lambda p: p.pdgid  # noqa: E731

    pdgids_baryons_defined_P = [
        pdgid(b)
        for b in Particle.findall(
            lambda p: p.P != Parity.u and p.pdgid.is_baryon and p.pdgid > 0
        )
    ]

    pdgids_baryons_undefined_P = [
        pdgid(b)
        for b in Particle.findall(
            lambda p: p.P == Parity.u and p.pdgid.is_baryon and p.pdgid > 0
        )
    ]

    for pdgid in pdgids_baryons_defined_P:
        assert Particle.from_pdgid(pdgid).P == -Particle.from_pdgid(-pdgid).P

    for pdgid in pdgids_baryons_undefined_P:
        assert Particle.from_pdgid(pdgid).P == Particle.from_pdgid(-pdgid).P


def test_C_consistency():
    """
    The charge conjugation parity is stored in the (curated) data CSV files.
    For neutral unflavoured mesons it can be calculated as C = (-1)^(L+S),
    and this relation can be checked against the CSV data.

    Note: mesons with PDGIDs of the kind 9XXXXXX (N=9) are not experimentally
    well-known particles and C is undefined.
    """
    for p in Particle.all():
        if not (p.is_unflavoured_meson and p.three_charge == 0):
            continue
        elif _digit(p.pdgid, Location.N) == 9:
            continue
        elif p.pdgid == 22:  # Special case of the photon
            assert p.C == -1
        elif p.pdgid in [130, 310]:  # Special case of the KS and KL
            assert p.C == Parity.u
        else:
            assert p.C == (-1) ** (p.L + p.S)


checklist_describe = (
    # Test undefined width value
    [1, "Width = None"],  # d quark
    # Test print-out of zero width values
    [22, "Width = 0.0 MeV"],  # photon
    # Test print-out of symmetric width errors
    [413, "Width = 0.0834 ± 0.0018 MeV"],  # D*(2010)+
    [443, "Width = 0.0926 ± 0.0017 MeV"],  # J/psi
    # Test print-out of asymmetric width errors
    [4222, "Width = 1.89 + 0.09 - 0.18 MeV"],  # Sigma_c(2455)++
    [23, "Width = 2495.2 ± 2.3 MeV"],  # H0
    # Test print-out of symmetric lifetime errors
    [5332, "Lifetime = 1.65e-03 + 1.8e-04 - 1.8e-04 ns"],  # Omega_b-
    [211, "Lifetime = 26.033 ± 0.005 ns"],  # pion
    # Test print-out of asymmetric lifetime errors
    [4332, "Lifetime = 2.7e-04 + 3e-05 - 3e-05 ns"],  # Omega_c^0
    # Test particles with at present an upper limit on their width
    [423, "Width < 2.1 MeV"],  # D*(2007)0
    [10431, "Width < 10.0 MeV"],  # D(s0)*(2317)+
    [20433, "Width < 6.3 MeV"],  # D(s1)(2460)+
)


@pytest.mark.parametrize("pid,description", checklist_describe)
def test_describe(pid, description):
    part = Particle.from_pdgid(pid)
    assert description in part.describe()


def test_default_table_loading():
    assert Particle.table_names() == ("particle2022.csv", "nuclei2020.csv")


def test_default_table_loading_bis():
    Particle.all()
    p = Particle.from_pdgid(211)
    assert p.table_loaded() is True
    assert p.table_names() == ("particle2022.csv", "nuclei2020.csv")


def test_explicit_table_loading():
    Particle.load_table(data.basepath / "particle2022.csv")
    assert Particle.table_loaded()
    assert len(Particle.table_names()) == 1
    assert Particle.all() is not None


def test_all_particles_are_loaded():
    Particle.load_table(data.basepath / "particle2018.csv")
    assert len(Particle.all()) == 605
    Particle.load_table(data.basepath / "particle2019.csv")
    assert len(Particle.all()) == 610
    Particle.load_table(data.basepath / "particle2020.csv")
    assert len(Particle.all()) == 610
    Particle.load_table(data.basepath / "particle2021.csv")
    assert len(Particle.all()) == 616
    Particle.load_table(data.basepath / "particle2022.csv")
    assert len(Particle.all()) == 616

    Particle.load_table(data.basepath / "nuclei2020.csv")
    assert len(Particle.all()) == 5880

    # Load default table to restore global state
    Particle.load_table()


checklist_html_name = (
    (22, "&#x03b3;"),  # photon
    (1, "d"),  # d quark
    (-2, "u&#773;"),  # u antiquark
    (11, "e<SUP>-</SUP>"),  # e-
    (-13, "&#x03bc;<SUP>+</SUP>"),  # mu+
    (-14, "&#x03bd;&#773;<SUB>&#x03bc;</SUB>"),  # nu_mu_bar
    (111, "&#x03c0;<SUP>0</SUP>"),  # pi0
    (-211, "&#x03c0;<SUP>-</SUP>"),  # pi-
    (-213, "&#x03c1;(770)<SUP>-</SUP>"),  # rho(770)-
    (20213, "a<SUB>1</SUB>(1260)<SUP>+</SUP>"),  # a_1(1260)+
    (321, "K<SUP>+</SUP>"),  # K+
    (130, "K<SUB>L</SUB><SUP>0</SUP>"),  # K_L
    (10321, "K<SUB>0</SUB><SUP>*</SUP>(1430)<SUP>+</SUP>"),  # K(0)*(1430)+
    (-10321, "K<SUB>0</SUB><SUP>*</SUP>(1430)<SUP>-</SUP>"),  # K(0)*(1430)-
    (10433, "D<SUB>s1</SUB>(2536)<SUP>+</SUP>"),  # D_s1(2536)+
    (-511, "B&#773;<SUP>0</SUP>"),  # B0_bar
    (443, "J/&#x03c8;(1S)"),  # J/psi
    (10441, "&#x03c7;<SUB>c0</SUB>(1P)"),  # chi_c0(1P)
    (300553, "&#x03a5;(4S)"),  # Upsilon(4S)
    (2212, "p"),  # proton
    (-2112, "n&#773;"),  # antineutron
    (-2224, "&#x0394;&#773;(1232)<SUP>--</SUP>"),  # Delta_bar(1232)--
    (3322, "&#x039e;<SUP>0</SUP>"),  # Xi0
    (-3322, "&#x039e;&#773;<SUP>0</SUP>"),  # Xi0_bar
    (-5122, "&#x039b;&#773;<SUB>b</SUB><SUP>0</SUP>"),  # Lb0_bar
)


@pytest.mark.parametrize("pid,html_name", checklist_html_name)
def test_html_name(pid, html_name):
    particle = Particle.from_pdgid(pid)

    assert particle.html_name == html_name


checklist_is_self_conjugate = (
    (1, False),  # d quark
    (-13, False),  # mu+
    (111, True),  # pi0
    (211, False),  # pi+
    (-211, False),  # pi-
    (443, True),  # J/psi
    (300553, True),  # Upsilon(4S)
    (130, True),  # K_L
    (2212, False),  # proton
    (-2112, False),  # antineutron
    (3322, False),  # Xi0
    (-3322, False),  # Xi0_bar
    (-511, False),  # B0_bar
    (5122, False),  # Lb0
)


@pytest.mark.parametrize("pid,is_self_conjugate", checklist_is_self_conjugate)
def test_is_self_conjugate(pid, is_self_conjugate):
    particle = Particle.from_pdgid(pid)

    assert particle.is_self_conjugate == is_self_conjugate


def test_self_conjugation_consistenty():
    """
    The logic implemented in ``Particle.invert()`` and ``Particle.is_self_conjugate``
    should be consistent. In other words, the inverse of
    ``self.anti_flag == Inv.ChargeInv and self.three_charge != Charge.o``
    in ``Particle.invert()`` should match ``Particle.is_self_conjugate``.
    """
    n_inconsistencies = sum(
        (p.anti_flag == Inv.ChargeInv and p.three_charge == Charge.o)
        and not p.is_self_conjugate
        for p in Particle.all()
    )

    assert n_inconsistencies == 0


checklist_is_name_barred = (
    (1, False),  # d quark
    (-2, True),  # u antiquark
    (11, False),  # e-
    (-13, False),  # mu+
    (111, False),  # pi0
    (211, False),  # pi+
    (-211, False),  # pi-
    (-213, False),  # rho(770)-
    (443, False),  # J/psi
    (300553, False),  # Upsilon(4S)
    (130, False),  # K_L
    (2212, False),  # proton
    (-2112, True),  # antineutron
    (3322, False),  # Xi0
    (-3322, True),  # Xi0_bar
    (-511, True),  # B0_bar
    (-5122, True),  # Lb0_bar
)


@pytest.mark.parametrize("pid,has_bar", checklist_is_name_barred)
def test_is_name_barred(pid, has_bar):
    particle = Particle.from_pdgid(pid)

    assert particle.is_name_barred == has_bar


def test_is_unflavoured_meson(PDGIDs):
    _unflavoured_mesons = (
        PDGIDs.Pi0,
        PDGIDs.PiPlus,
        PDGIDs.eta,
        PDGIDs.eta_prime,
        PDGIDs.a_0_1450_plus,
        PDGIDs.rho_770_minus,
        PDGIDs.phi,
        PDGIDs.omega,
        PDGIDs.rho_1700_0,
        PDGIDs.a2_1320_minus,
        PDGIDs.omega_3_1670,
        PDGIDs.f_4_2300,
        PDGIDs.jpsi,
        PDGIDs.psi_2S,
        PDGIDs.Upsilon_1S,
        PDGIDs.Upsilon_4S,
    )
    _non_unflavoured_mesons = [pid for pid in PDGIDs if pid not in _unflavoured_mesons]
    for pid in _unflavoured_mesons:
        try:
            assert Particle.from_pdgid(pid).is_unflavoured_meson
        except (ParticleNotFound, InvalidParticle):
            pass
    for pid in _non_unflavoured_mesons:
        try:
            assert not Particle.from_pdgid(pid).is_unflavoured_meson
        except (ParticleNotFound, InvalidParticle):
            pass


spin_type_classification = (
    # Gauge bosons
    (23, SpinType.Unknown),  # Z0 - no parity defined for it
    # Leptons aren't assigned a SpinType
    (11, SpinType.NonDefined),  # e-
    # Only mesons are given a SpinType
    # - Pseudo-scalars J^P = 0^-
    (211, SpinType.PseudoScalar),  # pi+
    (310, SpinType.PseudoScalar),  # K_S
    (-421, SpinType.PseudoScalar),  # D0_bar
    # - Scalars J^P = 0^+
    (9000211, SpinType.Scalar),  # a_0(980)+
    (9010221, SpinType.Scalar),  # f_0(980)
    # - Vector J^P = 1^-
    (333, SpinType.Vector),  # phi(1020)
    (443, SpinType.Vector),  # J/psi
    # Axial-vector - J^P = 1^+
    (20213, SpinType.Axial),  # a_1(1260)+
    (20313, SpinType.Axial),  # K_1(1400)0
    (10433, SpinType.Axial),  # D_s1(2536)+
    # Tensor - J^P = 2^+
    (225, SpinType.Tensor),  # f_2(1270)
    (315, SpinType.Tensor),  # K*_2(1430)0
    # Pseudo-tensor - J^P = 2^-
    (10225, SpinType.PseudoTensor),  # eta_2(1645)
    # J > 2 mesons
    (329, SpinType.Unknown),  # K*_4(2045)+
    # Baryons aren't assigned a SpinType
    (2212, SpinType.NonDefined),  # proton
)


@pytest.mark.parametrize("pid,stype", spin_type_classification)
def test_spin_type(pid, stype):
    particle = Particle.from_pdgid(pid)

    assert particle.spin_type == stype


checklist_isospin = (
    # Quarks
    (1, 0.5),  # d
    # Gauge bosons
    (22, None),  # photon
    (23, None),  # Z0
    # Leptons
    (11, None),  # e-
    (-12, None),  # nu(e)_bar
    # Mesons
    (211, 1.0),  # pi+
    (310, 0.5),  # K_S
    (-421, 0.5),  # D0_bar
    (333, 0.0),  # phi(1020)
    (443, 0.0),  # J/psi
    (521, 0.5),  # B+
    (531, 0.0),  # Bs
    # Baryons
    (2212, 0.5),  # proton
    (2214, 1.5),  # Delta+
)


@pytest.mark.parametrize("pid,isospin", checklist_isospin)
def test_isospin(pid, isospin):
    particle = Particle.from_pdgid(pid)

    assert particle.I == isospin  # noqa: E741


def test_default_particle():
    p = Particle.empty()

    assert repr(p) == '<Particle: name="Unknown", pdgid=0, mass=None>'
    assert "Name: Unknown" in p.describe()
    assert p.mass is None
    assert p.width is None
    assert p.spin_type == SpinType.NonDefined
    assert p.programmatic_name == "Unknown"
    assert p.status == Status.NotInPDT


def test_to_list():
    tbl = Particle.to_list(
        filter_fn=lambda p: p.pdgid.is_meson
        and p.pdgid.has_strange
        and p.ctau > 1 * meter,
        exclusive_fields=["pdgid", "name"],
    )
    assert tbl == [["pdgid", "name"], [130, "K(L)0"], [321, "K+"], [-321, "K-"]]

    tbl = Particle.to_list(
        filter_fn=lambda p: p.pdgid > 0
        and p.pdgid.is_meson
        and p.pdgid.has_strange
        and p.pdgid.has_charm,
        exclusive_fields=["name"],
        n_rows=2,
    )
    assert ["D(s)+"] in tbl
    assert ["D(s)*+"] in tbl


def test_to_dict():
    query_as_dict = Particle.to_dict(
        filter_fn=lambda p: p.pdgid.is_lepton and p.charge != 0,
        exclusive_fields=["name", "charge"],
        particle=False,
    )

    assert set(query_as_dict["name"]) == {"e+", "mu+", "tau+", "tau'+"}


ampgen_style_names = (
    ("b", 5),
    ("b~", -5),
    ("pi+", 211),
    ("pi-", -211),
    ("K~*0", -313),
    ("K*(892)bar0", -313),
    ("a(1)(1260)+", 20213),
    ("rho(1450)0", 100113),
    ("rho(770)0", 113),
    ("K(1)(1270)bar-", -10323),
    # ("K(1460)bar-", -100321),
    ("K(2)*(1430)bar-", -325),
)


@pytest.mark.parametrize("name,pid", ampgen_style_names)
def test_ampgen_style_names(name, pid):
    particle = Particle.from_string(name)

    assert particle.pdgid == pid
    assert particle == pid


decfile_style_names = (
    ("s", 3),
    ("anti-b", -5),
    ("anti-K*0", -313),
    ("eta", 221),
    ("eta'", 331),
    ("a_0+", 9000211),
    ("a_00", 9000111),
    ("a_1-", -20213),
    ("a_10", 20113),
    ("f_0", 9010221),
    ("f'_0", 10221),
    ("f_1", 20223),
    ("f'_1", 20333),
    ("f'_2", 335),
    ("h_1", 10223),
    ("h'_1", 10333),
    ("rho+", 213),
    ("rho(2S)0", 100113),
    ("omega", 223),
    ("omega(1650)", 30223),
    ("Delta++", 2224),
    ("Delta+", 2214),
    ("Delta0", 2114),
    ("Delta-", 1114),
    ("D+", 411),
    # ("D'_1+", 10413),
    # ("anti-D'_10", -10423),
    ("D_2*+", 415),
    ("D_s+", 431),
    ("anti-B0", -511),
    ("B+", 521),
    ("B-", -521),
    ("B*+", 523),
    ("B*-", -523),
    ("N(1440)+", 12212),
    ("anti-N(1440)-", -12212),
    ("anti-Lambda_b0", -5122),
    ("Sigma_b+", 5222),
    # ("Sigma_b0", 5212),
    ("Sigma_b-", 5112),
    ("anti-Sigma_b-", -5222),
    # ("anti-Sigma_b0", -5212),
    ("anti-Sigma_b+", -5112),
    ("Sigma_b*+", 5224),
    # ("Sigma_b*0", 5214),
    ("Sigma_b*-", 5114),
    ("anti-Sigma_b*-", -5224),
    # ("anti-Sigma_b*0", -5214),
    ("anti-Sigma_b*+", -5114),
)


@pytest.mark.parametrize("name,pid", decfile_style_names)
def test_decfile_style_names(name, pid):
    assert Particle.from_evtgen_name(name).pdgid == pid



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

# Python standard library
import contextlib
import csv
from copy import copy
from functools import total_ordering
from typing import Any, Callable, Iterable, Iterator, Sequence, SupportsInt, TypeVar

# External dependencies
import attr
from hepunits.constants import c_light

from .. import data
from ..converters.evtgen import EvtGenName2PDGIDBiMap
from ..pdgid import PDGID, is_valid
from ..pdgid.functions import Location, _digit
from ..typing import HasOpen, HasRead, StringOrIO, Traversable
from .enums import (
    Charge,
    Charge_mapping,
    Charge_undo,
    Inv,
    Parity,
    Parity_undo,
    SpinType,
    Status,
)
from .kinematics import width_to_lifetime
from .regex import getname
from .utilities import latex_to_html_name, programmatic_name, str_with_unc

__all__ = ("Particle", "ParticleNotFound", "InvalidParticle")


def __dir__() -> tuple[str, ...]:
    return __all__


class ParticleNotFound(RuntimeError):
    pass


class InvalidParticle(RuntimeError):
    pass


def _isospin_converter(isospin: str) -> float | None:
    vals: dict[str | None, float | None] = {
        "0": 0.0,
        "1/2": 0.5,
        "1": 1.0,
        "3/2": 1.5,
    }
    return vals.get(isospin)


def _none_or_positive_converter(value: float) -> float | None:
    return None if value < 0 else value


# These are needed to trick attrs typing
minus_one: float | None = -1.0
none_float: float | None = None


Self = TypeVar("Self", bound="Particle")


@total_ordering
@attr.s(slots=True, eq=False, order=False, repr=False)
class Particle:
    """
    The Particle object class. Hold a series of properties for a particle.

    Class properties:

    C
        The charge conjugation parity quantum number, if relevant.
        It is C = (-1)^(L+S) for self-conjugate mesons.
        Mesons with PDG IDs of the kind 9XXXXXX (N=9) are not experimentally well-known particles
        and None is returned.

    G
        The G-parity quantum number, if relevant.

    I
        The isospin quantum number, if relevant.

    P
        The parity quantum number, if relevant.
        It is P = (-1)^(L+1) for self-conjugate mesons and -1 for the photon.
        Mesons with PDG IDs of the kind 9XXXXXX (N=9) are not experimentally well-known particles
        and None is returned.

    anti_flag
        The particle-antiparticle flag.

        A = B     - particle that has anti-particle partner different from particle
                    with ASCII name formed by concatenation of the name shown below with charge
                    ( e- <--> e+, pi+ <--> pi-, K+ <--> K-, W+ <--> W- ).
        A = F     - particle that has anti-particle partner different from particle
                    with ascii name formed by concatenation of the name shown below with string "bar" and charge
                    by the rule (nu(e) <--> nubar(e), p <--> pbar, Delta++ <--> Deltabar--)
        A = blank - particle that coincides with its antiparticle (gamma, pi0, eta).

    charge
        The particle charge, in units of the positron charge.

    latex_name
        The particle name in LaTeX.

    mass
        The particle mass, in MeV.

    mass_lower
        The lower uncertainty on the particle mass, in MeV.

    mass_upper
        The upper uncertainty on the particle mass, in MeV.

    pdgid
        The PDG ID.

    pdg_name
        The particle name as in the PDG data file.

        Note:
        This name does not contain the charge. See alternative `name`.

    quarks
        The quark content of the particle. Empty string if not relevant.

        Note:
        Capital letters represent anti-quarks, 'qQ' stands for light quark,
        light anti-quark pair with unknown wave functions.
        x, y stand for wave function coefficients, see the Review of Particle Physics (RPP) 'Quark Model' Review.
        p, q stand for CP violation parameters, see the RPP 'CP violation in KL Decays' review.

    rank
        The particle rank as specified by the PDG, i.e. the number of baryon stars - used only on baryons.

        Possible values are:
        4 - Existence is certain, and properties are at least fairly well explored.
        3 - Existence ranges from very likely to certain, but further confirmation
             is desirable and/or quantum numbers, branching fractions, etc. are not well determined.
        2 - Evidence of existence is only fair.
        1 - Evidence of existence is poor.

    status
        The particle status as specified by the PDG.

        Possible values are:
        R - Established particles in the Review of Particle Physics (RPP) Summary Table
            in Particle Physics Booklet
            (established quarks, gauge bosons, leptons, mesons and baryons, except those in D below).
        D - The particle is omitted from the Summary Tables in the Particle Physics Booklet,
            but not from the Review. These entries are omitted only to save space
            even though they are well established.
        S - The particle is omitted from the particle properties Summary Tables
            because it is not well established.
        F - Special case: "Further mesons", see RPP, these states are in the RPP database
            but are poorly established or observed by a single group and thus need confirmation.
            If used, these should be referred to the original publication.

    width
        The particle decay width, in MeV.

    width_lower
        The lower uncertainty on the particle decay width, in MeV.

    width_upper
        The upper uncertainty on the particle decay width, in MeV.
    """

    pdgid: PDGID = attr.ib(converter=PDGID)
    pdg_name: str = attr.ib()
    mass: float | None = attr.ib(minus_one, converter=_none_or_positive_converter)
    mass_upper: float | None = attr.ib(minus_one, converter=_none_or_positive_converter)
    mass_lower: float | None = attr.ib(minus_one, converter=_none_or_positive_converter)
    width: float | None = attr.ib(minus_one, converter=_none_or_positive_converter)
    width_upper: float | None = attr.ib(
        minus_one, converter=_none_or_positive_converter
    )
    width_lower: float | None = attr.ib(
        minus_one, converter=_none_or_positive_converter
    )
    _three_charge: Charge | None = attr.ib(Charge.u, converter=Charge)  # charge * 3
    I: float | None = attr.ib(none_float, converter=_isospin_converter)  # noqa: E741
    # J = attr.ib(None)  # Total angular momentum
    G = attr.ib(Parity.u, converter=Parity)  # Parity: '', +, -, or ?
    P = attr.ib(Parity.u, converter=Parity)  # Space parity
    C = attr.ib(Parity.u, converter=Parity)  # Charge conjugation parity
    anti_flag = attr.ib(
        Inv.Same, converter=Inv
    )  # Info about particle name for anti-particles
    rank: int = attr.ib(0)
    status = attr.ib(Status.NotInPDT, converter=Status)
    quarks = attr.ib("", converter=str)
    latex_name: str = attr.ib("Unknown")

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: name="{self}", pdgid={int(self.pdgid)}, mass={self._str_mass()}>'

    # Ordered loaded table of entries
    _table: list[Particle] | None = None

    # Hash optimized table of particles
    _hash_table: dict[int, Particle] | None = None

    # Names of loaded tables
    _table_names: list[str] | None = None

    @classmethod
    def table_names(cls) -> tuple[str, ...]:
        """
        Return the list of names loaded.

        Note
        ----
        Calling this class method will load the default table,
        if no table has so far been loaded.
        Check with table_loaded() first if you don't want this loading
        to be triggered by the call.
        """

        if cls._table_names is None:
            cls.load_table()

        if cls._table_names is not None:
            return tuple(cls._table_names)  # make a copy to avoid user manipulation

        return ()

    @classmethod
    def table_loaded(cls) -> bool:
        """
        Check to see if the table is loaded.
        """
        return cls._table is not None

    @classmethod
    def all(cls: type[Self]) -> list[Self]:
        """
        Access, hence get hold of, the internal particle data CSV table,
        loading it from the default location if no table has yet been loaded.
        All `Particle` (instances) are returned as a list.
        """
        if not cls.table_loaded():
            cls.load_table()

        return cls._table if cls._table is not None else []  # type: ignore[return-value]

    @classmethod
    def to_list(
        cls,
        exclusive_fields: Iterable[str] = (),
        exclude_fields: Iterable[str] = (),
        n_rows: int = -1,
        filter_fn: Callable[[Particle], bool] | None = None,
        particle: bool | None = None,
        **search_terms: Any,
    ) -> Sequence[Sequence[bool | int | float | str]]:
        """
        Render a search (via `findall`) on the internal particle data CSV table
        as a `list`, loading the table from the default location if no table has yet been loaded.

        The returned attributes are those of the class. By default all attributes
        are used as fields. Their complete list is:
            pdgid
            pdg_name
            mass
            mass_upper
            mass_lower
            width
            width_upper
            width_lower
            three_charge
            I
            G
            P
            C
            anti_flag
            rank
            status
            quarks
            latex_name

        It is possible to add as returned fields any `Particle` class property,
        e.g. 'name', `J` or `ctau`, see examples below.

        Parameters
        ----------
        exclusive_fields: list, optional, default is []
            Exclusive list of fields to print out,
            which can be any `Particle` class property.
        exclude_fields: list, optional, default is []
            List of table fields to exclude in the printout.
            Relevant only when exclusive_fields is not given.
        n_rows: int, optional, defaults to all rows
            Number of table rows to print out.
        filter_fn: function, optional, default is None
            Apply a filter to each particle.
            See `findall(...)`` for typical use cases.
        particle: bool, optional, default is None
            Pass particle=True/False to only return particles or antiparticles.
            Option passed on internally to `findall(...)``.
        search_terms: keyword arguments, optional
            See `findall(...)`` for typical use cases.

        Returns
        -------
        The particle table query as a `list`.

        Note
        ----
        The `tabulate` package is suggested as a means to print-out
        the contents of the query as a nicely formatted table.

        Examples
        --------
        Reproduce the whole particle table kept internally:

        >>> query_as_list = Particle.to_list()

        Reduce the information on the particle table to the only fields
        ['pdgid', 'pdg_name'] and render the first 5 particles:

        >>> query_as_list = Particle.to_list(exclusive_fields=['pdgid', 'pdg_name'], n_rows=5)
        >>> from tabulate import tabulate
        >>> print(tabulate(query_as_list, headers='firstrow'))
          pdgid  pdg_name
        -------  ----------
              1  d
             -1  d
              2  u
             -2  u
              3  s

        Request the properties of a specific list of particles:

        >>> query_as_list = Particle.to_list(filter_fn=lambda p: p.pdgid.is_lepton and p.charge!=0, exclusive_fields=['pdgid', 'name', 'mass', 'charge'], particle=False)
        >>> print(tabulate(query_as_list, headers='firstrow', tablefmt="rst", floatfmt=".12g", numalign="decimal"))
        =======  ======  ===============  ========
          pdgid  name               mass    charge
        =======  ======  ===============  ========
            -11  e+         0.5109989461         1
            -13  mu+      105.6583745            1
            -15  tau+    1776.86                 1
            -17  tau'+                           1
        =======  ======  ===============  ========

        >>> query_as_list = Particle.to_list(filter_fn=lambda p: p.pdgid.is_lepton, pdg_name='tau', exclusive_fields=['pdgid', 'name', 'mass', 'charge'])
        >>> print(tabulate(query_as_list, headers='firstrow'))
          pdgid  name       mass    charge
        -------  ------  -------  --------
             15  tau-    1776.86        -1
            -15  tau+    1776.86         1

        Save it to a file:

        >>> with open('particles.txt', "w") as outfile:    # doctest: +SKIP
        ...    print(tabulate(query_as_list, headers='firstrow', tablefmt="rst", floatfmt=".12g", numalign="decimal"), file=outfile)    # doctest: +SKIP
        """
        if not cls.table_loaded():
            cls.load_table()

        # Get all table headers from the class attributes
        tbl_names = [a.name for a in attr.fields(Particle)]
        # ... and replace '_three_charge' with the better, public property
        tbl_names[tbl_names.index("_three_charge")] = "three_charge"

        list_exclusive_fields = list(exclusive_fields)
        if list_exclusive_fields:
            tbl_names = list_exclusive_fields
        else:
            for fld in exclude_fields:
                with contextlib.suppress(ValueError):
                    tbl_names.remove(fld)

        # Start with the full table
        tbl_all = sorted(cls.all())

        # Apply a filter, if specified
        if filter_fn is not None:
            tbl_all = cls.findall(filter_fn, particle=particle, **search_terms)

        # In any case, only keep a given number of rows?
        if n_rows >= 0:
            tbl_all = tbl_all[:n_rows]

        # Build all table rows
        tbl = [tbl_names]
        for p in tbl_all:
            tbl.append([getattr(p, attr) for attr in tbl_names])

        return tbl

    @classmethod
    def to_dict(
        cls, *args: Any, **kwargs: Any
    ) -> dict[str, tuple[bool | int | str | float, ...]]:
        """
        Render a search (via `findall`) on the internal particle data CSV table
        as a `dict`, loading the table from the default location if no table has yet been loaded.

        See `to_list` for details on the full function signature.

        The returned attributes are those of the class. By default all attributes
        are used as fields. Their complete list is:
            pdgid
            pdg_name
            mass
            mass_upper
            mass_lower
            width
            width_upper
            width_lower
            three_charge
            I
            G
            P
            C
            anti_flag
            rank
            status
            quarks
            latex_name

        It is possible to add as returned fields any `Particle` class property,
        e.g. 'name', `J` or `ctau`, see examples below.

        Parameters
        ----------
        exclusive_fields: list, optional, default is []
            Exclusive list of fields to print out,
            which can be any `Particle` class property.
        exclude_fields: list, optional, default is []
            List of table fields to exclude in the printout.
            Relevant only when exclusive_fields is not given.
        n_rows: int, optional, defaults to all rows
            Number of table rows to print out.
        filter_fn: function, optional, default is None
            Apply a filter to each particle.
            See `findall(...)`` for typical use cases.
        particle: bool, optional, default is None
            Pass particle=True/False to only return particles or antiparticles.
            Option passed on internally to `findall(...)``.
        search_terms: keyword arguments, optional
            See `findall(...)`` for typical use cases.

        Returns
        -------
        The particle table query as a `dict`.

        Note
        ----
        The `tabulate` package is suggested as a means to print-out
        the contents of the query as a nicely formatted table.

        Examples
        --------
        Reproduce the whole particle table kept internally:

        >>> query_as_dict = Particle.to_dict()

        Reduce the information on the particle table to the only fields
        ['pdgid', 'pdg_name'] and render the first 5 particles:

        >>> query_as_dict = Particle.to_dict(exclusive_fields=['pdgid', 'pdg_name'], n_rows=5)
        >>> from tabulate import tabulate    # doctest: +SKIP
        >>> print(tabulate(query_as_dict, headers='keys'))    # doctest: +SKIP
          pdgid  pdg_name
        -------  ----------
              1  d
             -1  d
              2  u
             -2  u
              3  s

        Request the properties of a specific list of particles:

        >>> query_as_dict = Particle.to_dict(filter_fn=lambda p: p.pdgid.is_lepton and p.charge!=0, exclusive_fields=['pdgid', 'name', 'mass', 'charge'], particle=True)
        >>> print(tabulate(query_as_dict, headers='keys', tablefmt="rst", floatfmt=".12g", numalign="decimal"))    # doctest: +SKIP
        =======  ======  ===============  ========
          pdgid  name               mass    charge
        =======  ======  ===============  ========
             11  e-         0.5109989461        -1
             13  mu-      105.6583745           -1
             15  tau-    1776.86                -1
             17  tau'-                          -1
        =======  ======  ===============  ========

        >>> query_as_dict = Particle.to_dict(filter_fn=lambda p: p.pdgid.is_lepton, pdg_name='tau', exclusive_fields=['pdgid', 'name', 'mass', 'charge'])
        >>> print(tabulate(query_as_dict, headers='keys'))    # doctest: +SKIP
          pdgid  name       mass    charge
        -------  ------  -------  --------
             15  tau-    1776.86        -1
            -15  tau+    1776.86         1

        Save it to a file:

        >>> with open('particles.txt', "w") as outfile:    # doctest: +SKIP
        ...    print(tabulate(query_as_dict, headers='keys', tablefmt="rst", floatfmt=".12g", numalign="decimal"), file=outfile)    # doctest: +SKIP
        """
        query_as_list = cls.to_list(*args, **kwargs)
        headers = query_as_list[0]
        rows = zip(*query_as_list[1:])

        return {str(h): r for h, r in zip(headers, rows)}

    @classmethod
    def load_table(
        cls,
        filename: StringOrIO | None = None,
        append: bool = False,
        _name: str | None = None,
    ) -> None:
        """
        Load a particle data CSV table. Optionally append to the existing data already loaded if append=True.
        As a special case, if this is called with append=True and the table is not loaded, the default will
        be loaded first before appending (set append=False if you don't want this behavior).

        A parameter is also included that should be considered private for now. It is _name, which
        will override the filename for the stored filename in _table_names.
        """

        if append and not cls.table_loaded():
            cls.load_table(append=False)  # default load
        elif not append:
            cls._table = []
            cls._hash_table = {}
            cls._table_names = []

        # Tell MyPy that this is true
        assert cls._table is not None
        assert cls._hash_table is not None
        assert cls._table_names is not None

        if filename is None:
            with data.basepath.joinpath("particle2022.csv").open() as fa:
                cls.load_table(fa, append=append, _name="particle2022.csv")
            with data.basepath.joinpath("nuclei2020.csv").open() as fb:
                cls.load_table(fb, append=True, _name="nuclei2020.csv")
            return
        if isinstance(filename, HasRead):
            tmp_name = _name or filename.name
            cls._table_names.append(tmp_name or f"{filename!r} {len(cls._table_names)}")
            open_file = filename
        elif isinstance(filename, HasOpen):
            cls._table_names.append(str(filename))
            open_file = filename.open()
        else:
            cls._table_names.append(str(filename))
            assert not isinstance(filename, Traversable)
            open_file = open(filename, encoding="utf_8")

        with open_file as f:
            r = csv.DictReader(line for line in f if not line.startswith("#"))

            for v in r:
                with contextlib.suppress(ValueError):
                    value = int(v["ID"])

                    p = cls(
                        pdgid=value,
                        mass=float(v["Mass"]),
                        mass_upper=float(v["MassUpper"]),
                        mass_lower=float(v["MassLower"]),
                        width=float(v["Width"]),
                        width_upper=float(v["WidthUpper"]),
                        width_lower=float(v["WidthLower"]),
                        I=v["I"],
                        G=int(v["G"]),
                        P=int(v["P"]),
                        C=int(v["C"]),
                        anti_flag=int(v["Anti"]),
                        three_charge=int(v["Charge"]),
                        rank=int(v["Rank"]),
                        status=int(v["Status"]),
                        pdg_name=v["Name"],
                        quarks=v["Quarks"],
                        latex_name=v["Latex"],
                    )

                    # Replace the previous value if it exists
                    cls._hash_table[value] = p

        cls._table = sorted(cls._hash_table.values())

    # The following __le__ and __eq__ needed for total ordering (sort, etc)

    def __lt__(self, other: Particle | int) -> bool:
        # Sort by absolute particle numbers
        # The positive one should come first
        if isinstance(other, Particle):
            return abs(int(self.pdgid) - 0.25) < abs(int(other.pdgid) - 0.25)
        # Comparison with anything else should produce normal comparisons.
        return int(self.pdgid) < other

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Particle):
            return self.pdgid == other.pdgid
        return self.pdgid == other

    # Only one particle can exist per PDGID number
    def __hash__(self) -> int:
        return hash(self.pdgid)

    # Shared with PDGID

    @property
    def J(self) -> int:
        """
        The total spin J quantum number.

        Note that the returned value corresponds to that effectively encoded
        in the particle PDG ID.
        """
        return self.pdgid.J  # type: ignore[no-any-return]

    @property
    def L(self) -> int | None:
        """
        The orbital angular momentum L quantum number (None if not a meson).

        Note that the returned value corresponds to that effectively encoded
        in the particle PDG ID.
        """
        return self.pdgid.L  # type: ignore[no-any-return]

    @property
    def S(self) -> int | None:
        """
        The spin S quantum number (None if not a meson).

        Note that the returned value corresponds to that effectively encoded
        in the particle PDG ID.
        """
        return self.pdgid.S  # type: ignore[no-any-return]

    @property
    def charge(self) -> float | None:
        """
        The particle charge, in units of the positron charge.

        Design note: the particle charge can also be retrieved from the particle PDG ID.
        To allow for user-defined particles, it is necessary to rather store
        the particle charge in the data tables themselves.
        Consistency of both ways of retrieving the particle charge is guaranteed
        for all PDG table particles.
        """
        return self.three_charge / 3 if self.three_charge is not None else None

    @property
    def three_charge(self) -> int | None:
        "Three times the particle charge (charge * 3), in units of the positron charge."
        if not self.pdgid.is_nucleus:
            # Return int(...) not to return the actual enum Charge
            return (
                int(self._three_charge)
                if self._three_charge is not None and self._three_charge != Charge.u
                else None
            )

        return self.pdgid.three_charge  # type: ignore[no-any-return]

    @property
    def lifetime(self) -> float | None:
        """
        The particle lifetime, in nanoseconds.

        None is returned if the particle width (stored in the DB) is unknown.
        """
        return width_to_lifetime(self.width) if self.width is not None else None

    @property
    def ctau(self) -> float | None:
        """
        The particle c*tau, in millimeters.

        None is returned if the particle width (stored in the DB) is unknown.
        """
        return (
            c_light * self.lifetime
            if self.width is not None and self.lifetime is not None
            else None
        )

    @property
    def is_name_barred(self) -> bool:
        """
        Check to see if particle is inverted (hence is it an antiparticle)
        and has a bar in its name.
        """
        return self.pdgid < 0 and self.anti_flag == Inv.Barred

    @property
    def spin_type(self) -> SpinType:
        """
        Access the SpinType enum.

        Note that this is relevant for bosons only. SpinType.NonDefined is returned otherwise.
        """
        # Non-valid or non-standard PDG IDs
        if self.pdgid.j_spin is None:
            return SpinType.NonDefined

        # Fermions - 2J+1 is always an even number
        if self.pdgid.j_spin % 2 == 0:
            return SpinType.NonDefined

        J = int(self.J)
        if J in {0, 1, 2}:
            if self.P == Parity.p:
                return (SpinType.Scalar, SpinType.Axial, SpinType.Tensor)[J]
            if self.P == Parity.m:
                spin_types = (
                    SpinType.PseudoScalar,
                    SpinType.Vector,
                    SpinType.PseudoTensor,
                )
                return spin_types[J]

        return SpinType.Unknown

    @property
    def is_self_conjugate(self) -> bool:
        """
        Is the particle self-conjugate, i.e. its own antiparticle?
        """
        return self.anti_flag == Inv.Same

    @property
    def is_unflavoured_meson(self) -> bool:
        """
        Is the particle a light non-strange meson or a quarkonium?

        Indeed, unflavoured mesons are either:
        all light mesons with no net flavour quantum number (S = C = B = T = 0),
        or quarkonia, which are self-conjugate heavy flavour mesons,
        with all their flavour quantum numbers and charge equal to zero.
        """
        pid = self.pdgid

        if not pid.is_meson:
            return False

        # Heavy flavour
        if pid.has_charm or pid.has_bottom or pid.has_top:
            return self.is_self_conjugate

        # Light or strange mesons at this point

        # Special case of the KS and KL
        if pid in {130, 310}:
            return False

        # I = 1 light mesons have no s-sbar component, hence has_strange == False
        if _digit(pid, Location.Nq3) == 1 and not pid.has_strange:
            return True

        # I = 0 light mesons have a s-sbar component, has_strange == True,
        # thought their net S = 0
        if _digit(pid, Location.Nq3) in {2, 3} and self.three_charge == 0:
            return True

        # Only K-mesons at this point
        return False

    def invert(self: Self) -> Self:
        "Get the antiparticle."
        if self.anti_flag == Inv.Barred or (
            self.anti_flag == Inv.ChargeInv and self.three_charge != Charge.o
        ):
            return self.from_pdgid(-self.pdgid)
        return copy(self)

    __neg__ = invert
    __invert__ = invert

    # Pretty descriptions

    def __str__(self) -> str:
        _tilde = "~" if self.anti_flag == Inv.Barred and self.pdgid < 0 else ""
        _charge = self._str_charge() if self._charge_in_name() else ""
        return self.pdg_name + _tilde + _charge

    name = property(
        __str__,
        doc="The nice name, with charge added, and a tilde for an antiparticle, if relevant.",
    )

    def _repr_latex_(self) -> str:
        name = self.latex_name
        return ("$" + name + "$") if self.latex_name else "?"

    def _width_or_lifetime(self) -> str:
        """
        Display either the particle width or the lifetime.
        Internally used by the describe() method.

        Note
        ----
        Width errors equal to None flag an experimental upper limit on the width.
        """
        if self.width is None:
            return "Width = None"
        if self.width == 0:
            return "Width = 0.0 MeV"
        if self.width_lower is None or self.width_upper is None:
            return f"Width < {self.width} MeV"
        if (
            self.width < 0.05
        ):  # corresponds to a lifetime of approximately 1.3e-20 seconds
            assert self.lifetime is not None
            if self.width_lower != self.width_upper:
                lifetime = str_with_unc(
                    self.lifetime,
                    width_to_lifetime(self.width - self.width_lower) - self.lifetime,
                    self.lifetime - width_to_lifetime(self.width + self.width_upper),
                )
                return f"Lifetime = {lifetime} ns"

            e = width_to_lifetime(self.width - self.width_lower) - self.lifetime
            lifetime = str_with_unc(self.lifetime, e, e)
            return f"Lifetime = {lifetime} ns"

        width = str_with_unc(self.width, self.width_upper, self.width_lower)
        return f"Width = {width} MeV"

    def _charge_in_name(self) -> bool:
        """Assess whether the particle charge is part of the particle name.

        Internally used when creating the name.
        """
        if self.anti_flag == Inv.ChargeInv:
            return True  # antiparticle flips sign of particle
        if self.pdgid in (23, 25, 111, 130, 310, 311, -311):
            return True  # the Z0, H0, pi0, KL0, KS0, K0 and K0bar
        if self.pdgid.is_diquark:
            return False
        if abs(self.pdgid) in (2212, 2112):
            return False  # proton and neutron
        if abs(self.pdgid) < 19:
            return False  # all quarks and neutrinos (charged leptons dealt with in 1st line of if statements ;-))
        if self.three_charge is None:
            return False  # deal with corner cases ;-)
        if self.is_self_conjugate:
            pid = self.pdgid
            if pid < 25:
                return False  # Gauge bosons
            # Quarkonia never exhibit the 0 charge
            # All eta, eta', h, h', omega, phi, f, f' light mesons are supposed to have an s-sbar component (see PDG site),
            # but some particles have pdgid.has_strange==False :S! Play it safe ...
            if any(
                chr in self.pdg_name
                for chr in ("eta", "h(", "h'(", "omega", "phi", "f", "f'")
            ):
                return False
            if pid.has_strange or pid.has_charm or pid.has_bottom or pid.has_top:
                return False

            # Light unflavoured mesons
            return True
        # Lambda baryons
        if (
            self.pdgid.is_baryon
            and _digit(self.pdgid, Location.Nq2) == 1
            and self.I
            == 0.0  # 1st check alone is not sufficient to filter out lowest-ground Sigma's
            and self.pdgid.has_strange
            and not (
                self.pdgid.has_charm or self.pdgid.has_bottom or self.pdgid.has_top
            )
        ):
            return False
        if self.pdgid.is_nucleus:
            return False
        return True

    def _str_charge(self) -> str:
        """
        Display a reasonable particle charge printout.
        Internally used by the describe() and __str__ methods.
        """
        if self._three_charge is None:
            return "None"
        if not self.pdgid.is_nucleus:
            return Charge_undo[Charge(self._three_charge)]

        return str(self.pdgid.charge)

    def _str_mass(self) -> str:
        """
        Display a reasonable particle mass printout
        even when no mass value is available.
        Internally used by the describe() method.
        """
        if self.mass is None:
            return "None"

        txt = str_with_unc(self.mass, self.mass_upper, self.mass_lower)
        return f"{txt} MeV"

    def describe(self) -> str:
        "Make a nice high-density string for a particle's properties."

        if self.pdgid == 0:
            return "Name: Unknown"

        G = Parity_undo[self.G]
        C = Parity_undo[self.C]
        Q = self._str_charge()
        P = Parity_undo[self.P]
        mass = self._str_mass()
        width_or_lifetime = self._width_or_lifetime()
        latex_name = self._repr_latex_()

        val = f"""Name: {self!s:<14} ID: {self.pdgid:<12} Latex: {latex_name}
Mass  = {mass}
{width_or_lifetime}
Q (charge)        = {Q:<6}  J (total angular) = {self.J!s:<7}  P (space parity) = {P}
C (charge parity) = {C:<6}  I (isospin)       = {self.I!s:<7}  G (G-parity)     = {G}
"""

        if self.spin_type != SpinType.Unknown:
            val += f"    SpinType: {self.spin_type!s}\n"
        if self.quarks:
            val += f"    Quarks: {self.quarks}\n"
        val += f"    Antiparticle name: {self.invert().name} (antiparticle status: {self.anti_flag.name})"
        return val

    @property
    def evtgen_name(self) -> str:
        "This is the name used in EvtGen."
        return EvtGenName2PDGIDBiMap[self.pdgid]

    @property
    def programmatic_name(self) -> str:
        "This name could be used for a variable name."
        return programmatic_name(self.name)

    @property
    def html_name(self) -> str:
        "This is the name in HTML."
        return latex_to_html_name(self.latex_name)

    @classmethod
    def empty(cls: type[Self]) -> Self:
        "Make a new empty particle."
        return cls(0, "Unknown", anti_flag=Inv.Same)

    @classmethod
    def from_pdgid(cls: type[Self], value: SupportsInt) -> Self:
        """
        Get a particle from a PDGID. Uses by default the package
        extended PDG data table.

        Raises
        ------
        InvalidParticle
            If the input PDG ID is an invalid identification code.
        ParticleNotFound
            If no matching PDG ID is found in the loaded data table(s).
        """
        if not is_valid(value):
            raise InvalidParticle(f"Input PDGID {value} is invalid!")

        if not cls.table_loaded():
            cls.load_table()

        assert cls._hash_table is not None

        try:
            return cls._hash_table[int(value)]  # type: ignore[return-value]
        except KeyError:
            raise ParticleNotFound(f"Could not find PDGID {value}") from None

    @classmethod
    def from_name(cls: type[Self], name: str) -> Self:
        """
        Get a particle from its name.

        Raises
        ------
        ParticleNotFound
            If no particle matches the input name.
        """
        try:
            (particle,) = cls.finditer(
                name=name
            )  # throws an error if < 1 or > 1 particle is found
            return particle
        except ValueError:
            raise ParticleNotFound(f'Could not find name "{name}"') from None

    @classmethod
    def from_evtgen_name(cls: type[Self], name: str) -> Self:
        """
        Get a particle from an EvtGen particle name, as in .dec decay files.

        Raises
        ------
        ParticleNotFound
            If `from_pdgid`, internally called, returns no match.
        MatchingIDNotFound
            If the matching EvtGen name - PDG ID done internally is unsuccessful.
        """
        return cls.from_pdgid(EvtGenName2PDGIDBiMap[name])

    @classmethod
    def from_nucleus_info(
        cls: type[Self],
        z: int,
        a: int,
        anti: bool = False,
        i: int = 0,
        l_strange: int = 0,
    ) -> Self:
        """
        Get a nucleus particle from the proton Z and atomic mass A numbers.

        As described in the PDG numbering scheme:
        "To avoid ambiguities, nucleus codes should not be applied to a single hadron, such as the p, n
        or the Λ, where quark-contents-based codes already exist."

        Number of neutrons is equal to a-z.
        The PDG ID format is ±10LZZZAAAI, see the `is_nucleus()` function in submodule
        `particle.pdgid.functions.py`.

        Parameters
        ----------
        z: int
            Atomic number Z (number of protons).
            Maximum three decimal digits.
        a: int
            Atomic mass number A (total number of baryons).
            Maximum three decimal digits.
        anti: bool, optional, defaults to False (not antimatter).
            Whether the nucleus should be antimatter.
        i: int, optional, default is 0
            Isomer level I. I=0 corresponds to the ground-state.
            I>0 are excitations.
            Maximum is one decimal digit.
        l_strange: int, optional, default is 0
            Total number of strange quarks.
            Maximum is one decimal digit.

        Raises
        ------
        ParticleNotFound
            If `from_pdgid`, internally called, returns no match.
        InvalidParticle
            If the input combination is invalid.
        """
        if l_strange < 0 or l_strange > 9:
            raise InvalidParticle(
                f"Number of strange quarks l={l_strange} is invalid. Must be 0 <= l <= 9."
            )
        if z < 0 or z > 999:
            raise InvalidParticle(
                f"Atomic number Z={z} is invalid. Must be 0 <= A <= 999."
            )
        if a < 0 or a > 999:
            raise InvalidParticle(
                f"Atomic mass number A={a} is invalid. Must be 0 <= A <= 999."
            )
        if i < 0 or i > 9:
            raise InvalidParticle(
                f"Isomer level I={i} is invalid. Must be 0 <= I <= 9."
            )
        if z > a:
            raise InvalidParticle(
                f"Nucleus A={a}, Z={z} is invalid. Z must be smaller or equal to Z."
            )

        pdgid = int(1e9 + l_strange * 1e5 + z * 1e4 + a * 10 + i)

        if anti:
            return cls.from_pdgid(-pdgid)

        return cls.from_pdgid(pdgid)

    @classmethod
    def finditer(
        cls: type[Self],
        filter_fn: Callable[[Particle], bool] | str | None = None,
        *,
        particle: bool | None = None,
        **search_terms: Any,
    ) -> Iterator[Self]:
        """
        Search for a particle, returning an iterator of candidates.

        The first and only positional argument is given to each particle
        candidate, and returns True/False. Example:

            >>> Particle.finditer(lambda p: 'p' in p.name)    # doctest: +SKIP
            # Returns iterator of all particles with p somewhere in name

        You can pass particle=True/False to force a particle or antiparticle.
        If this is not callable, it will do a "fuzzy" search on the name. So this is identical:

            >>> Particle.finditer('p')    # doctest: +SKIP
            # Returns literator of all particles with p somewhere in name (same as example above)

        You can also pass keyword arguments, which are either called with the
        matching property if they are callable, or are compared if they are not.
        This would do an exact search on the name, instead of a fuzzy search:

           >>> # Returns proton and antiproton only
           >>> Particle.finditer(pdg_name='p')    # doctest: +SKIP
           # Returns iterator

           >>> # Returns proton only
           >>> Particle.finditer(pdg_name='p', particle=True)    # doctest: +SKIP
           # Returns iterator

        Versatile searches require a (lambda) function as argument:

        >>> # Get all neutral beauty hadrons
        >>> Particle.finditer(lambda p: p.pdgid.has_bottom and p.charge==0)    # doctest: +SKIP
        >>>
        >>> # Trivially find all pseudoscalar charm mesons
        >>> Particle.finditer(lambda p: p.pdgid.is_meson and p.pdgid.has_charm and p.spin_type==SpinType.PseudoScalar)  # doctest: +SKIP

        See also ``findall``, which returns the same thing, but as a list.
        """

        # Filter out values
        for item in cls.all():
            # At this point, continue if a match fails

            # particle=True is particle, False is antiparticle, and None is both
            if particle is not None:
                if particle and int(item.pdgid) < 0:
                    continue
                if (not particle) and int(item.pdgid) > 0:
                    continue

            # If a filter function is passed, evaluate and skip if False
            if filter_fn is not None:
                if callable(filter_fn):
                    # Just skip exceptions, which are there for good reasons
                    # Example: calls to Particle.ctau for particles given
                    #          default negative, and hence invalid, widths
                    try:
                        if not filter_fn(item):
                            continue
                    except TypeError:  # skip checks such as 'lambda p: p.width > 0',
                        continue  # which fail when width=None
                elif filter_fn not in item.name:
                    continue

            # At this point, if you break, you will not add a match
            for term, value in search_terms.items():
                # If pvalue cannot be accessed, skip this particle
                # (invalid lifetime, for example)
                try:
                    pvalue = getattr(item, term)
                except ValueError:
                    break

                # Callables are supported
                if callable(value):
                    try:
                        if not value(pvalue):
                            break
                    except TypeError:  # catch issues with None values
                        break
                # And, finally, just compare if nothing else matched
                elif pvalue != value:
                    break

            # If the loop was not broken
            else:
                yield item

    @classmethod
    def findall(
        cls: type[Self],
        filter_fn: Callable[[Particle], bool] | None = None,
        *,
        particle: bool | None = None,
        **search_terms: Any,
    ) -> list[Self]:

        """
        Search for a particle, returning a list of candidates.

        The first and only positional argument is given to each particle
        candidate, and returns True/False. Example:

            >>> Particle.findall(lambda p: 'p' in p.name)    # doctest: +SKIP
            # Returns list of all particles with p somewhere in name

        You can pass particle=True/False to force a particle or antiparticle.
        If this is not callable, it will do a "fuzzy" search on the name. So this is identical:

            >>> Particle.findall('p')    # doctest: +SKIP
            # Returns list of all particles with p somewhere in name (same as example above)

        You can also pass keyword arguments, which are either called with the
        matching property if they are callable, or are compared if they are not.
        This would do an exact search on the name, instead of a fuzzy search:

           >>> # Returns proton and antiproton only
           >>> Particle.findall(pdg_name='p')    # doctest: +NORMALIZE_WHITESPACE
           [<Particle: name="p", pdgid=2212, mass=938.272081 ± 0.000006 MeV>,
            <Particle: name="p~", pdgid=-2212, mass=938.272081 ± 0.000006 MeV>,
            <Particle: name="p", pdgid=1000010010, mass=938.272081 ± 0.000006 MeV>,
            <Particle: name="p~", pdgid=-1000010010, mass=938.272081 ± 0.000006 MeV>]

           >>> # Returns proton only
           >>> Particle.findall(pdg_name='p', particle=True)    # doctest: +NORMALIZE_WHITESPACE
           [<Particle: name="p", pdgid=2212, mass=938.272081 ± 0.000006 MeV>,
           <Particle: name="p", pdgid=1000010010, mass=938.272081 ± 0.000006 MeV>]

        Versatile searches require a (lambda) function as argument:

        >>> # Get all neutral beauty hadrons
        >>> Particle.findall(lambda p: p.pdgid.has_bottom and p.charge==0)    # doctest: +SKIP
        >>>
        >>> # Trivially find all pseudoscalar charm mesons
        >>> Particle.findall(lambda p: p.pdgid.is_meson and p.pdgid.has_charm and p.spin_type==SpinType.PseudoScalar)  # doctest: +SKIP

        See also ``finditer``, which provides an iterator instead of a list.
        """

        return list(
            cls.finditer(filter_fn=filter_fn, particle=particle, **search_terms)
        )

    @classmethod
    def find(cls: type[Self], *args: Any, **search_terms: Any) -> Self:
        raise AttributeError(
            "Method has been removed post versions 0.16. Please use finditer or findall instead."
        )

    @classmethod
    def from_string(cls: type[Self], name: str) -> Self:
        "Get a particle from a PDG style name - returns the best match."
        matches = cls.from_string_list(name)
        if matches:
            return matches[0]
        raise ParticleNotFound(f"{name} not found in particle table")

    @classmethod
    def from_string_list(cls: type[Self], name: str) -> list[Self]:
        "Get a list of particles from a PDG style name."

        # Forcible override
        particle = None

        short_name = name
        if "~" in name:
            short_name = name.replace("~", "")
            particle = False

        # Try the simplest searches first
        list_can = cls.findall(name=name, particle=particle)
        if list_can:
            return list_can
        list_can = cls.findall(pdg_name=short_name, particle=particle)
        if list_can:
            return list_can

        mat_str = getname.match(short_name)

        if mat_str is None:
            return []

        mat = mat_str.groupdict()

        if particle is False:
            mat["bar"] = "bar"

        try:
            return cls._from_group_dict_list(mat)
        except ParticleNotFound:
            return []

    @classmethod
    def _from_group_dict_list(cls: type[Self], mat: dict[str, Any]) -> list[Self]:

        kw: dict[str, Any] = {
            "particle": False
            if mat["bar"] is not None
            else True
            if mat["charge"] == "0"
            else None
        }

        name = mat["name"]

        if mat["family"]:
            if "_" in mat["family"]:
                mat["family"] = mat["family"].strip("_")
            name += f'({mat["family"]})'
        if mat["state"]:
            name += f'({mat["state"]})'

        if "prime" in mat and mat["prime"]:
            name += "'"

        if mat["star"]:
            name += "*"

        if mat["state"] is not None:
            kw["J"] = float(mat["state"])

        maxname = name + f'({mat["mass"]})' if mat["mass"] else name
        if "charge" in mat and mat["charge"] is not None:
            kw["three_charge"] = Charge_mapping[mat["charge"]]

        vals = cls.findall(name=lambda x: maxname in x, **kw)
        if not vals:
            vals = cls.findall(name=lambda x: name in x, **kw)

        if not vals:
            raise ParticleNotFound(f"Could not find particle {maxname} or {name}")

        if len(vals) > 1 and mat["mass"] is not None:
            vals = [val for val in vals if mat["mass"] in val.latex_name]

        if len(vals) > 1:
            vals = sorted(vals)

        return vals



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

import sys

from .corsika import Corsika7ID
from .geant import Geant3ID

# Direct access to Particle literals
# Direct access to Particle (the CSV file is not read until a particle is accessed)
from .particle import InvalidParticle, Particle, ParticleNotFound, literals
from .particle.enums import Charge, Inv, Parity, SpinType, Status

# Direct access to PDGID and other ID classes
from .pdgid import PDGID
from .pythia import PythiaID

# Convenient access to the version number
from .version import version as __version__

sys.modules["particle.literals"] = literals

# Direct access to handy LaTeX to HTML particle name conversions
# Direct access to kinematics functions
from .particle import latex_to_html_name, lifetime_to_width, width_to_lifetime

__all__ = (
    "Charge",
    "Corsika7ID",
    "Geant3ID",
    "Inv",
    "InvalidParticle",
    "PDGID",
    "Parity",
    "Particle",
    "ParticleNotFound",
    "PythiaID",
    "SpinType",
    "Status",
    "latex_to_html_name",
    "lifetime_to_width",
    "width_to_lifetime",
    "__version__",
)


def __dir__() -> tuple[str, ...]:
    return __all__



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

import pytest

from particle import Particle
from particle.lhcb import from_lhcb_name, to_lhcb_name

lhcb_style_names = (
    ("nu_tau", 16),
    ("nu_tau~", -16),
    ("eta_prime", 331),
    ("f'_2(1525)", 335),
    ("D*_s+", 433),
    ("D*_s-", -433),
    ("B_s0", 531),
    ("B_s~0", -531),
    ("Lambda_b0", 5122),
    ("Lambda_b~0", -5122),
    # ("X_1(3872)", 9920443),
)


@pytest.mark.parametrize("name,pid", lhcb_style_names)
def test_from_lhcb_name(name, pid):
    assert from_lhcb_name(name).pdgid == pid


@pytest.mark.parametrize("name,pid", lhcb_style_names)
def test_to_lhcb_name(name, pid):
    assert to_lhcb_name(Particle.from_pdgid(pid)) == name



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

import pytest

from particle import data
from particle.converters.bimap import BiMap, DirectionalMaps
from particle.exceptions import MatchingIDNotFound
from particle.pdgid import PDGID
from particle.pythia import PythiaID


def test_BiMap():
    bimap = BiMap(PDGID, PythiaID)

    assert len(bimap) == 540
    assert "BiMap(PDGID-PythiaID)" in str(bimap)

    with pytest.raises(MatchingIDNotFound):
        bimap[PDGID(9000221)]


def test_DirectionalMaps():
    filename = data.basepath / "pdgid_to_pythiaid.csv"
    PDG2PyIDMap, Py2PDGIDMap = DirectionalMaps(
        "PDGID", "PythiaID", filename=filename, converters=(int, int)
    )

    assert len(PDG2PyIDMap) == 540
    assert len(Py2PDGIDMap) == 540

    assert "DirectionalMap(PDGID->PYTHIAID)" in str(PDG2PyIDMap)
    assert "DirectionalMap(PYTHIAID->PDGID)" in str(Py2PDGIDMap)

    with pytest.raises(MatchingIDNotFound):
        PDG2PyIDMap[PDGID(9000221)]
    with pytest.raises(MatchingIDNotFound):
        Py2PDGIDMap[PythiaID(9000221)]



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

from ..pdgid import PDGID
from ..pythia import PythiaID
from .bimap import BiMap

Pythia2PDGIDBiMap = BiMap(PDGID, PythiaID)
Pythia2PDGIDBiMap.__doc__ = """
Bi-bidirectional map between PDG and Pythia IDs.

Examples
--------
>>> pyid = Pythia2PDGIDBiMap[PDGID(9010221)]
>>> pyid
<PythiaID: 10221>

>>> pdgid = Pythia2PDGIDBiMap[PythiaID(10221)]
>>> pdgid
<PDGID: 9010221>
"""



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

import pytest
from hepunits.constants import hbar
from hepunits.units import GeV, MeV, ps
from pytest import approx

from particle.particle import lifetime_to_width, width_to_lifetime


def test_valid_width_lifetime_conversions():
    assert lifetime_to_width(1.5 * ps) / GeV == approx(4.388079676311604e-13)
    assert 1.5 * ps * lifetime_to_width(1.5 * ps) == hbar
    assert width_to_lifetime(hbar) == 1 * MeV


def test_invalid_width_lifetime_conversions():
    with pytest.raises(ValueError):
        lifetime_to_width(-1)
    with pytest.raises(ValueError):
        width_to_lifetime(-1)

    assert lifetime_to_width(0) == float("inf")
    assert width_to_lifetime(0) == float("inf")



#!/usr/bin/env python
# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.
"""
Script to generate the pdgid_to_corsika7id.csv conversion table from Corsika7ID to PDGID and vice-versa.
This script should be kept, so the table won't need to be hand-edited in the future.
"""
from __future__ import annotations

import csv
import datetime as dt
import pathlib

from particle import Particle, ParticleNotFound
from particle.shared_literals import common_particles

# Pairs of matching (Corsika7ID, PDGID),
# if the Corsika7ID has a matching PDGID, if not
# then (Corsika7ID, str), with the string from
# the Corsika7 user guide
corsica_pdg_id = [
    (1, common_particles["gamma"]),
    (50, common_particles["omega_782"]),
    (2, common_particles["e_plus"]),
    (51, common_particles["rho_770_0"]),
    (3, common_particles["e_minus"]),
    (52, common_particles["rho_770_plus"]),
    (53, common_particles["rho_770_minus"]),
    (5, common_particles["mu_plus"]),
    (54, common_particles["Delta_1232_pp"]),
    (6, common_particles["mu_minus"]),
    (55, common_particles["Delta_1232_plus"]),
    (7, common_particles["pi_0"]),
    (56, common_particles["Delta_1232_0"]),
    (8, common_particles["pi_plus"]),
    (57, common_particles["Delta_1232_minus"]),
    (9, common_particles["pi_minus"]),
    (58, common_particles["Delta_1232_mm_bar"]),
    (10, common_particles["K_L_0"]),
    (59, common_particles["Delta_1232_minus_bar"]),
    (11, common_particles["K_plus"]),
    (60, common_particles["Delta_1232_0_bar"]),
    (12, common_particles["K_minus"]),
    (61, common_particles["Delta_1232_plus_bar"]),
    (13, common_particles["neutron"]),
    (62, common_particles["Kst_892_0"]),
    (14, common_particles["proton"]),
    (63, common_particles["Kst_892_plus"]),
    (15, common_particles["antiproton"]),
    (64, common_particles["Kst_892_minus"]),
    (16, common_particles["K_S_0"]),
    (65, common_particles["Kst_892_0_bar"]),
    (17, common_particles["eta"]),
    (66, common_particles["nu_e"]),
    (18, common_particles["Lambda"]),
    (67, common_particles["nu_e_bar"]),
    (19, common_particles["Sigma_plus"]),
    (68, common_particles["nu_mu"]),
    (20, common_particles["Sigma_0"]),
    (69, common_particles["nu_mu_bar"]),
    (21, common_particles["Sigma_minus"]),
    (22, common_particles["Xi_0"]),
    (71, "η → γγ"),
    (23, common_particles["Xi_minus"]),
    (72, "η → 3π◦"),
    (24, common_particles["Omega_minus"]),
    (73, "η → π+π−π◦"),
    (25, common_particles["antineutron"]),
    (74, "η → π+π−γ"),
    (26, common_particles["Lambda_bar"]),
    (75, "μ+ add. info."),
    (27, common_particles["Sigma_minus_bar"]),
    (76, "μ− add. info."),
    (28, common_particles["Sigma_0_bar"]),
    (29, common_particles["Sigma_plus_bar"]),
    (85, "decaying μ+ at start"),
    (30, common_particles["Xi_0_bar"]),
    (86, "decaying μ− at start"),
    (31, common_particles["Xi_plus_bar"]),
    (32, common_particles["Omega_plus_bar"]),
    (95, "decaying μ+ at end"),
    (48, common_particles["etap_958"]),
    (49, common_particles["phi_1020"]),
    (96, "decaying μ− at end"),
    (116, common_particles["D_0"]),
    (155, common_particles["Xi_cp_minus_bar"]),
    (117, common_particles["D_plus"]),
    (156, common_particles["Xi_cp_0_bar"]),
    (118, common_particles["D_minus"]),
    (157, common_particles["Omega_c_0_bar"]),
    (119, common_particles["D_0_bar"]),
    (120, common_particles["D_s_plus"]),
    (161, common_particles["Sigma_c_2455_pp"]),
    (121, common_particles["D_s_minus"]),
    (162, common_particles["Sigma_c_2455_plus"]),
    (122, common_particles["eta_c_1S"]),
    (163, common_particles["Sigma_c_2455_0"]),
    (123, common_particles["Dst_2007_0"]),
    (124, common_particles["Dst_2010_plus"]),
    (171, common_particles["Sigma_c_2455_mm_bar"]),
    (125, common_particles["Dst_2010_minus"]),
    (172, common_particles["Sigma_c_2455_minus_bar"]),
    (126, common_particles["Dst_2007_0_bar"]),
    (173, common_particles["Sigma_c_2455_0_bar"]),
    (127, common_particles["D_sst_plus"]),
    (128, common_particles["D_sst_minus"]),
    (176, common_particles["B_0"]),
    (177, common_particles["B_plus"]),
    (130, common_particles["Jpsi_1S"]),
    (178, common_particles["B_minus"]),
    (131, common_particles["tau_plus"]),
    (179, common_particles["B_0_bar"]),
    (132, common_particles["tau_minus"]),
    (180, common_particles["B_s_0"]),
    (133, common_particles["nu_tau"]),
    (181, common_particles["B_s_0_bar"]),
    (134, common_particles["nu_tau_bar"]),
    (182, common_particles["B_c_plus"]),
    (183, common_particles["B_c_minus"]),
    (137, common_particles["Lambda_c_plus"]),
    (184, common_particles["Lambda_b_0"]),
    (138, common_particles["Xi_c_plus"]),
    (185, common_particles["Sigma_b_minus"]),
    (139, common_particles["Xi_c_0"]),
    (186, common_particles["Sigma_b_plus"]),
    (140, common_particles["Sigma_c_2455_pp"]),
    (187, common_particles["Xi_b_0"]),
    (141, common_particles["Sigma_c_2455_plus"]),
    (188, common_particles["Xi_b_minus"]),
    (142, common_particles["Sigma_c_2455_0"]),
    (189, common_particles["Omega_b_minus"]),
    (143, common_particles["Xi_cp_plus"]),
    (190, common_particles["Lambda_b_0_bar"]),
    (144, common_particles["Xi_cp_0"]),
    (191, common_particles["Sigma_b_plus_bar"]),
    (145, common_particles["Sigma_c_2455_0"]),
    (192, common_particles["Sigma_b_minus_bar"]),
    (193, common_particles["Xi_b_0_bar"]),
    (149, common_particles["Lambda_c_minus_bar"]),
    (194, common_particles["Xi_b_plus_bar"]),
    (150, common_particles["Xi_c_minus_bar"]),
    (195, common_particles["Omega_b_plus_bar"]),
    (151, common_particles["Xi_c_0_bar"]),
    (152, common_particles["Sigma_c_2455_mm_bar"]),
    (153, common_particles["Sigma_c_2455_minus_bar"]),
    (154, common_particles["Sigma_c_2455_0_bar"]),
]


def dump_pdgid_to_corsika7(file: pathlib.Path | None = None) -> None:
    """
    Generates the conversion .csv file with the patching PDGID to Corsika7ID under 'src/particle/data/pdgid_to_corsika7id.csv'
    (if file is None, else in the specified path).
    """
    # Loop over all thinkable values and only add them if the PDG ID exists
    for a in range(2, 56 + 1):
        for z in range(0, a + 1):
            corsikaid = a * 100 + z
            try:
                corsica_pdg_id.append(
                    (corsikaid, int(Particle.from_nucleus_info(a=a, z=z).pdgid))
                )
            except ParticleNotFound:
                pass

    if file is None:
        file = (
            pathlib.Path(__file__)
            .parent.parent.resolve()
            .absolute()
            .joinpath("src/particle/data/pdgid_to_corsika7id.csv")
        )

    date = dt.datetime.today().strftime("%Y-%m-%d")

    with open(
        file,
        "w",
        newline="",
        encoding="utf-8",
    ) as csvfile:
        csvfile.write(
            f"# (c) Scikit-HEP project - Particle package data file - pdgid_to_corsika7id.csv - {date}\n"
        )
        csvfile.write(
            "# Auto generated by 'admin/dump_pdgid_to_corsika7.py'\n",
        )

        writer = csv.writer(csvfile)
        # Header
        writer.writerow(("PDGID", "CORSIKA7ID"))
        for corsikaid, pdgid in sorted(corsica_pdg_id, key=lambda x: x[0]):
            if isinstance(pdgid, int):
                writer.writerow((pdgid, corsikaid))


if __name__ == "__main__":
    dump_pdgid_to_corsika7()



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
Definitions of handy particle `Particle` literals
=================================================

The particle literals provide aliases for most common particles, with easily recognisable names.
The aliases are instances of the Particle class.

Typical use cases::

    >>> from particle import literals as lp
    >>> lp.pi_plus
    <Particle: name="pi+", pdgid=211, mass=139.57039 ± 0.00018 MeV>
    >>> lp.pi_plus.name
    'pi+'
    >>> from particle.literals import Lambda_b_0
    >>> Lambda_b_0
    <Particle: name="Lambda(b)0", pdgid=5122, mass=5619.60 ± 0.17 MeV>
    >>> Lambda_b_0.J
    0.5

List of available/defined literals:

{0}
"""

from __future__ import annotations

from ..shared_literals import common_particles
from .particle import Particle


def __dir__() -> list[str]:
    return list(common_particles)


for item in common_particles:
    locals()[item] = Particle.from_pdgid(common_particles[item])


__doc = "".join(
    f"  {item} = Particle.from_pdgid({common_particles[item]})\n"
    for item in common_particles
)
__doc__ = __doc__.format(__doc)



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

from .converters import LHCbName2PDGIDBiMap
from .functions import from_lhcb_name, to_lhcb_name

__all__ = (
    "from_lhcb_name",
    "to_lhcb_name",
    "LHCbName2PDGIDBiMap",
)


def __dir__() -> tuple[str, ...]:
    return __all__



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

import pytest

# Requires pandas
pd = pytest.importorskip("pandas")

from collections import Counter

from particle import data
from particle.particle.convert import produce_files

FILES = ["particle2021.csv", "particle2022.csv"]


def test_generate(tmp_path):
    "This verifies that the input and output files match."

    particle2021 = tmp_path / "particle2021.csv"
    particle2022 = tmp_path / "particle2022.csv"

    produce_files(particle2021, particle2022, "DUMMY", "2022")

    """
    # No longer test this file, which eventually will be removed
    particle2018_data = data.basepath / "particle2018.csv"
    with particle2018.open() as src, particle2018_data as res:
        src = [l for l in src.readlines() if not l.startswith("#")]
        res = [l for l in res.readlines() if not l.startswith("#")]
        assert src == res
    """

    particle2022_data = data.basepath / "particle2022.csv"
    with particle2022.open() as src, particle2022_data.open() as res:
        src = [line for line in src.readlines() if not line.startswith("#")]
        res = [line for line in res.readlines() if not line.startswith("#")]
        assert src == res


@pytest.mark.parametrize("filename", FILES)
def test_csv_file_duplicates(filename):
    particle_data = data.basepath / filename
    p = pd.read_csv(particle_data, comment="#")

    duplicates = {item for item, count in Counter(p.ID).items() if count > 1}
    assert duplicates == set()


@pytest.mark.parametrize("filename", FILES)
def test_csv_file_has_latex(filename):
    particle_data = data.basepath / filename
    p = pd.read_csv(particle_data, comment="#")

    assert p[p.Latex == ""].empty



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

from .. import data
from ..pdgid import PDGID
from .bimap import BiMap, DirectionalMaps

EvtGenName2PDGIDBiMap = BiMap(
    PDGID,
    str,
    converters=(int, str),
    filename=data.basepath / "pdgid_to_evtgenname.csv",
)
EvtGenName2PDGIDBiMap.__doc__ = """
Bi-bidirectional map between PDG IDs and EvtGen names.

Examples
--------
>>> name = EvtGenName2PDGIDBiMap[PDGID(22)]
>>> name
'gamma'

>>> pdgid = EvtGenName2PDGIDBiMap['gamma']
>>> pdgid
<PDGID: 22>
"""


PDG2EvtGenNameMap, EvtGen2PDGNameMap = DirectionalMaps("PDGName", "EvtGenName")

PDG2EvtGenNameMap.__doc__ = """
Directional map between PDG and EvtGen names.

Examples
--------
>>> PDG2EvtGenNameMap['J/psi(1S)']
'J/psi'
"""


EvtGen2PDGNameMap.__doc__ = """
Directional map between EvtGen and names.

Examples
--------
>>> EvtGen2PDGNameMap['J/psi']
>>> 'J/psi(1S)'
"""

del BiMap
del DirectionalMaps



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
Collection of regular expression helper utilities for the ``Particle`` class.
"""

from __future__ import annotations

import re

getname = re.compile(
    r"""
^                                           # Beginning of string
      (?P<name>       \w+?        )         # One or more characters, non-greedy
(?:\( (?P<family>    [udsctb][\w]*) \) )?   # Optional family like (s)
(?:\( (?P<state>      \d+         ) \)      # Optional state in ()
      (?=             \*? \(      )  )?     #   - lookahead for mass
      (?P<star>       \*          )?        # Optional star
(?:\( (?P<mass>       \d+         ) \) )?   # Optional mass in ()
      (?P<bar>        (bar|~)     )?        # Optional bar
      (?P<charge>     [0\+\-][+-]?)         # Required 0, -, --, or +, ++
$                                           # End of string
""",
    re.VERBOSE,
)


# Help manipulating .dec DecFile style names
getdec = re.compile(
    r"""
^                                           # Beginning of string
      (?P<bar>        (anti-)     )?        # Optional anti-
      (?P<name>       [a-zA-Z]+?  )         # One or more characters, non-greedy
      (?P<prime>      '*          )         # Optional prime(s)
(?: _ (?P<state>      \d+         )    )?   # Optional state in ()
(?:   (?P<family>     _[udsctbemna][\w]*) )?# Optional family like (s)
(?:\( (?P<mass>       \d+         ) \) )?   # Optional mass in ()
      (?P<star>       \*?         )         # Optional star
      (?P<charge>     [0\+\-][+-]?)?        # Optional 0, -, --, or +, ++
$                                           # End of string
""",
    re.VERBOSE,
)



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

from particle.pdgid import PDGID
from particle.pdgid import functions as _functions
from particle.pdgid.pdgid import _fnames


def test_class_string_representations():
    pid = PDGID(11)
    assert pid == 11
    assert pid.__str__() == "<PDGID: 11>"
    pid = PDGID(-99999999)
    assert pid.__str__() == "<PDGID: -99999999 (is_valid==False)>"


def test_class_operations(PDGIDs):
    id_electron = PDGID(PDGIDs.Electron)
    id_positron = PDGID(PDGIDs.Positron)
    assert PDGIDs.Electron == id_electron
    assert id_positron == -id_electron
    assert PDGIDs.Positron == -id_electron


def test_class_return_type():
    assert isinstance(-PDGID(311), PDGID)
    assert isinstance(~PDGID(311), PDGID)


def test_class_inversion():
    assert -PDGID(311) == ~PDGID(311)


def test_nonphysical_pdgids():
    # The negative PDGID of a self-conjugate meson makes no sense
    assert not (
        PDGID(-111).is_meson
    )  # the "anti-pi0" with opposite PDGID of a pi0 does not exist
    assert not (
        PDGID(-443).is_meson
    )  # the "anti-J/psi" with opposite PDGID of a J/psi does not exist


def test_info():
    __info = """A              None
J              1.0
L              None
S              None
Z              None
abspid         22
charge         0.0
has_bottom     False
has_charm      False
has_down       False
has_fundamental_anti False
has_strange    False
has_top        False
has_up         False
is_Qball       False
is_Rhadron     False
is_SUSY        False
is_baryon      False
is_diquark     False
is_dyon        False
is_excited_quark_or_lepton False
is_gauge_boson_or_higgs True
is_generator_specific False
is_hadron      False
is_lepton      False
is_meson       False
is_nucleus     False
is_pentaquark  False
is_quark       False
is_sm_gauge_boson_or_higgs True
is_sm_lepton   False
is_sm_quark    False
is_special_particle False
is_technicolor False
is_valid       True
j_spin         3
l_spin         None
s_spin         None
three_charge   0
"""
    assert PDGID(22).info() == __info


def test_decorated_class_methods(PDGIDs):
    """
    Check that all particle.pdgid functions decorated in the PDGID class
    work as expected for all kinds of PDGIDs.
    """
    for m in _fnames:
        for pid in PDGIDs:
            assert getattr(PDGID(pid), m) == getattr(_functions, m)(pid)



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
Helper (internal) module with particle aliases
for all particles in the loaded "database" CSV file, excluding nuclei.

See the particle.literals and the pdgid.literals submodules for the actually exposed aliases.
"""


from __future__ import annotations

from .particle import Particle

# Make aliases for all particles in the latest "database", excluding nuclei
common_particles = {
    p.programmatic_name: int(p.pdgid)
    for p in Particle.findall(lambda p: abs(p.pdgid) < 1000000000)
}

# Some extra names that are expected:
common_particles.update(
    photon=22,  # official programmatic name is "gamma"
    Higgs=25,  # official programmatic name is "H_0"
    proton=2212,  # official programmatic name is "p"
    antiproton=-2212,  # official programmatic name is "p_bar"
    neutron=2112,  # official programmatic name is "n"
    antineutron=-2112,  # official programmatic name is "n_bar"
)



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
Class representing a Corsika7 ID.

Note
----
Corsika8 uses Geant3 Particle IDs.
"""


from __future__ import annotations

import csv
from typing import TypeVar

from .. import data
from ..exceptions import MatchingIDNotFound
from ..pdgid import PDGID

Self = TypeVar("Self", bound="Corsika7ID")


with data.basepath.joinpath("pdgid_to_corsika7id.csv").open() as _f:
    _bimap = {
        int(v["CORSIKA7ID"]): int(v["PDGID"])
        for v in csv.DictReader(line for line in _f if not line.startswith("#"))
    }

# Some Corsika7 ID's are not really particles
_non_particles = {
    71: "η → γγ",
    72: "η → 3π◦",
    73: "η → π+π−π◦",
    74: "η → π+π−γ",
    75: "μ+ add. info.",
    76: "μ− add. info.",
    85: "decaying μ+ at start",
    86: "decaying μ− at start",
    95: "decaying μ+ at end",
    96: "decaying μ− at end",
}


class Corsika7ID(int):
    """
    Holds a Corsika7 ID.

    All arithmetic operations on the class (like `-Corsika7ID(5)`) will
    return an integer. This is unlike for example `-PDGID(13)`, which will
    return `PDGID(-13)`. But since negative values have no direct meaning
    as a Corsika7ID, (in the output file they are used to indicate mother-particles)
    we omit this feature.

    Examples
    --------
    >>> cid = Corsika7ID(6)

    >>> from particle import Particle
    >>> p = Particle.from_pdgid(cid.to_pdgid())

    >>> (p,) = Particle.finditer(pdgid=cid.to_pdgid())
    >>> p.name
    'mu-'
    """

    __slots__ = ()  # Keep Corsika7ID a slots based class

    @classmethod
    def from_pdgid(cls: type[Self], pdgid: int) -> Self:
        """
        Constructor from a PDGID.
        """
        for k, v in _bimap.items():
            if v == pdgid:
                return cls(k)
        raise MatchingIDNotFound(f"Non-existent Corsika7ID for input PDGID {pdgid}!")

    @classmethod
    def from_particle_description(
        cls: type[Self], particle_description: int
    ) -> tuple[Self, bool]:
        """
        Constructor from the particle description returned by Corsika7
        in the particle data sub-block, mother particle data sub-block or
        the grandmother particle data sub-block.

        Returns
        -------
        A tuple with

        Corsika7ID: The Corsika7 id
        bool:       If the particle is a (grand)mother particle.
        """
        cid = abs(particle_description) // 1000
        ismother = particle_description < 0

        if cls._is_non_particle_id(cid):
            return cls(cid), ismother

        # This catches the cases of nuclei with no known PDG ID
        if 200 <= cid < 5699:
            return cls(cid), ismother

        if cid in _bimap:
            return cls(cid), ismother

        raise MatchingIDNotFound(
            f"Non-existent Corsika7ID for particle description {particle_description}!"
        )

    @classmethod
    def _is_non_particle_id(cls: type[Self], corsikaid: int) -> bool:
        """
        Returns True if the ID is valid but does not correspond to a particle, False otherwise.
        """
        return (
            corsikaid in _non_particles
            or corsikaid // 1000 == 8888
            or corsikaid == 9900
        )

    def is_particle(self) -> bool:
        """
        Returns True if the ID really belongs to a particle, since some are for example additional information.

        Examples
        --------
        >>> mu_minux = Corsika7ID(6)
        >>> mu_minus.is_particle()
        True
        >>> mu_info = Corsika7ID(76)
        >>> mu_info.is_particle()
        False
        >>> mu_info.name()
        'μ− add. info.'
        """
        iid = int(self)

        return not self._is_non_particle_id(iid)

    def name(self) -> str:
        """
        Returns a human readable name of the Corsika7ID.
        This also works for non-particles (is_particle()==false).

        Raises
        ------
        ParticleNotFound
            If it is a 'valid' PDG particle, but unknown.
            This for example happens with strange nuclei, which are not in the nuclei list.

        Examples
        --------
        >>> mu_minus = Corsika7ID(6)
        >>> mu_minus.is_particle()
        True
        >>> mu_minus.name() # For a particle, this returns the same name as `Particle.name`
        'mu'
        >>> mu_info = Corsika7ID(76)
        >>> mu_info.is_particle()
        False
        >>> mu_info.name()
        'μ− add. info.'
        >>> ch_photons_of = Corsika7ID(9900)
        >>> ch_photons_of.is_particle()
        False
        >>> ch_photons_of.name()
        'Cherenkov photons on particle output file'
        """
        from ..particle.particle import Particle  # pylint: disable=C0415

        if self.is_particle():
            return str(Particle.from_pdgid(self.to_pdgid()).name)

        iid = int(self)

        if iid in _non_particles:
            return _non_particles[iid]

        if iid // 1000 == 8888:
            return "weights of preceding particle (MULTITHIN option)"

        if iid == 9900:
            return "Cherenkov photons on particle output file"

        raise RuntimeError("This should be unreachable.")

    def to_pdgid(self) -> PDGID:
        """
        Raises
        ------
        InvalidParticle
            If it is a valid Corsika particle, but not a valid PDGID.

        Examples
        --------
        >>> Corsika7ID(6).to_pdgid()
        <PDGID: 13>
        >>> Corsika7ID(76).to_pdgid()
        InvalidParticle: The Corsika7ID <Corsika7ID: 76> does not correspond to a particle and thus has no equivalent PDGID.
        """
        from ..particle.particle import InvalidParticle  # pylint: disable=C0415

        if self not in _bimap:
            raise InvalidParticle(
                f"The Corsika7ID {self} does not correspond to a particle and thus has no equivalent PDGID."
            )
        return PDGID(_bimap[self])

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {int(self):d}>"

    def __str__(self) -> str:
        return repr(self)



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

"""
This is a conversion file, not part of the public API.

The default CSV files can be updated directly using the command:

    >>> python -m particle.particle.convert regenerate 2022 <version_number>    # doctest: +SKIP

A custom fwf file and LaTeX file can be converted into the CSV format using:

    >>> python -m particle.particle.convert extended output.csv file.fwf <version_number> latex.csv    # doctest: +SKIP

The command is typically

    >>> python -m particle.particle.convert extended output.csv src/particle/data/mass_width_2008.fwf <version_number>    # doctest: +SKIP

This file requires pandas. But most users will not need this file, as it only
converts PDG data files into the CSV file(s) the public API tools use. The tests
load some of these functions to verify the CSV files are in sync with the sources.

Internal usage
--------------

You can load a table from a classic "extended style" PDG table (only produced in 2008),
combined with one or more LaTeX files describing the pair (PDG ID, LaTeX name):

    >>> full_table = get_from_pdg_extended('particle/data/mass_width_2008.fwf',
    ...                                    ['particle/data/pdgid_to_latexname.csv'])

You can also read in a modern "standard" file (this will produce fewer columns):

    >>> ext_table = get_from_pdg_mcd('particle/data/mass_width_2022.mcd')

A utility is even provided to use the modern table to update the full table:

    >>> new_table = update_from_mcd(full_table, ext_table)

You can see what particles were missing from the full table if you want:

    >>> rem = set(ext_table.index) - set(full_table.index)
    >>> print(ext_table.loc[rem].sort_index())    # doctest: +SKIP

When you are done, you can save one or more of the tables:

    >>> full_table.to_csv('particle2008.csv', float_format='%.8g')

"""

from __future__ import annotations

import os
from datetime import date
from io import StringIO
from pathlib import Path
from typing import Any, Callable, Iterable, TextIO, TypeVar

import numpy as np
import pandas as pd

from .. import data
from ..pdgid import PDGID, is_baryon
from ..typing import StringOrIO, Traversable
from .enums import (
    Charge,
    Charge_mapping,
    Inv,
    Inv_mapping,
    Parity,
    Parity_mapping,
    Status_mapping,
)

FILE_DIR = os.path.dirname(os.path.realpath(__file__))

__all__ = (
    "convert",
    "get_from_pdg_extended",
    "get_from_pdg_mcd",
    "get_from_latex",
    "update_from_mcd",
)


def __dir__() -> tuple[str, ...]:
    return __all__


def get_from_latex(filename: StringOrIO) -> pd.Series:
    """
    Produce a pandas series from a file with LaTeX mappings in itself.
    The CVS file format is the following: PDGID, ParticleLatexName.
    """
    latex_table = pd.read_csv(filename, index_col=0, comment="#")
    return latex_table.LATEXNAME


def filter_file(fileobject: StringOrIO) -> TextIO:
    """
    Open a file if not already a file-like object, and strip lines that start with *.
    Returns a new file-like object (StringIO instance).
    """

    if not hasattr(fileobject, "read"):
        assert isinstance(fileobject, str)
        with open(fileobject, encoding="utf-8") as f:
            return filter_file(f)

    assert not isinstance(fileobject, (str, Traversable))

    stream = StringIO()
    for line in fileobject:
        # We need to strip the unicode byte ordering if present before checking for *
        if not line.lstrip("\ufeff").lstrip().startswith("*"):
            stream.write(line)
    stream.seek(0)

    return stream


T = TypeVar("T")


def get_from_pdg_extended(
    filename: StringOrIO, latexes: Iterable[StringOrIO] = ()
) -> pd.DataFrame:
    """
    Read an "extended style" PDG data file (only produced in 2008), plus a list of LaTeX files,
    to produce a pandas DataFrame with particle information.

    Parameters
    ----------
    filename: string
        Input file name
    latexes: list
        A list of names of LaTeX files describing the pair (PDG ID, LaTeX name) in CSV format

    Example
    -------
    >>> full_table = get_from_pdg_extended('particle/data/mass_width_2008.fwf',
    ...                                    ['particle/data/pdgid_to_latexname.csv'])
    """

    # Read a file, plus a list of LaTeX files, to produce a pandas DataFrame with particle information
    def unmap(mapping: dict[str, T]) -> Callable[[str], T]:
        return lambda x: mapping[x.strip()]

    # Convert each column from text to appropriate data type
    PDG_converters = dict(
        Charge=unmap(Charge_mapping),
        G=unmap(Parity_mapping),
        P=unmap(Parity_mapping),
        C=unmap(Parity_mapping),
        Anti=unmap(Inv_mapping),
        Rank=lambda x: int(x.strip()) if x.strip() else 0,
        ID=lambda x: int(x.strip()) if x.strip() else -1,
        Status=unmap(Status_mapping),
        Name=lambda x: x.strip(),
        I=lambda x: x.strip(),  # noqa: E741
        J=lambda x: x.strip(),
        Quarks=lambda x: x.strip(),
    )

    with filter_file(filename) as file_object:
        # Read in the table, apply the converters, add names, ignore comments
        pdg_table = pd.read_csv(
            file_object,
            names="Mass,MassUpper,MassLower,Width,WidthUpper,WidthLower,I,G,J,P,C,Anti,"
            "ID,Charge,Rank,Status,Name,Quarks".split(","),
            converters=PDG_converters,
            comment="#",
        )

    # Read the LaTeX
    latex_series = pd.concat([get_from_latex(latex) for latex in latexes])

    # Filtering out non-particles (quarks, negative IDs)
    # pdg_table = pdg_table[pdg_table.Charge != Charge.u]
    pdg_table = pdg_table[pdg_table.ID >= 0]

    # PDG's ID should be the key to table
    pdg_table.set_index("ID", inplace=True)

    # Assign the positive values LaTeX names
    pdg_table = pdg_table.assign(Latex=latex_series)

    # Some post processing to produce inverted particles
    pdg_table_inv = pdg_table[
        (pdg_table.Anti == Inv.Barred)
        | (
            (pdg_table.Anti == Inv.ChargeInv)
            # Maybe add?    & (pdg_table.Charge != Charge.u)
            & (pdg_table.Charge != Charge.o)
        )
    ].copy()

    pdg_table_inv.index = -pdg_table_inv.index
    pdg_table_inv.Charge = -pdg_table_inv.Charge
    pdg_table_inv.Quarks = (
        pdg_table_inv.Quarks.str.swapcase()
        .str.replace("SQRT", "sqrt")
        .str.replace("P", "p")
        .str.replace("Q", "q")
        .str.replace("mAYBE NON", "Maybe non")
        .str.replace("X", "x")
        .str.replace("Y", "y")
    )

    # Parity flips for baryons
    def is_baryon_with_defined_parity(i: PDGID, p: Parity) -> bool:
        return is_baryon(i) and p != Parity.u

    pdg_table_inv["P"] = np.where(
        pdg_table_inv.reset_index().apply(
            lambda x: is_baryon_with_defined_parity(x["ID"], x["P"]), axis=1
        ),
        -pdg_table_inv["P"],
        pdg_table_inv["P"],
    )

    full_inversion = pdg_table_inv.Anti == Inv.Barred
    pdg_table_inv.Latex.where(
        ~full_inversion,
        pdg_table_inv.Latex.str.replace(
            r"^(\\mathrm{|)([a-zA-Z\\][a-zA-Z]*)", r"\1\\overline{\2}", regex=True
        ),
        inplace=True,
    )
    pdg_table_inv.Latex = (
        pdg_table_inv.Latex.str.replace(r"+", r"%", regex=True)
        .str.replace(r"-", r"+")
        .str.replace(r"%", r"-")
    )

    # Make a combined table with + and - ID numbers
    full = pd.concat([pdg_table, pdg_table_inv])

    # This will override any negative values
    full.Latex.update(latex_series)

    # These items are not very important - can be reconstructed from the PDG ID
    # TODO: maybe first check the consistency between what is read in and what the PDG ID provides (being maniac)?
    del full["J"]

    # Nice sorting
    sort_particles(full)

    # All the 'MassLower' and 'WidthLower' values should be absolute values
    # except for the special cases when they are equal to -1,
    # which flag experimental upper limits or badly known particles
    full.loc[full["MassLower"] != -1, "MassLower"] = abs(full["MassLower"])
    full.loc[full["WidthLower"] != -1, "WidthLower"] = abs(full["WidthLower"])

    # Return the table, making sure NaNs are just empty strings, and sort
    return full.fillna("")


def sort_particles(table: pd.DataFrame) -> None:
    "Sort a particle list table nicely"
    table["TmpVals"] = abs(table.index - 0.25)
    table.sort_values("TmpVals", inplace=True)
    del table["TmpVals"]


def get_from_pdg_mcd(filename: StringOrIO) -> pd.DataFrame:
    """
    Reads in a current-style PDG .mcd file (mass_width_2022.mcd file tested).

    Example
    -------
    >>> mcd_table = get_from_pdg_mcd('particle/data/mass_width_2022.mcd')
    """

    # The format here includes the space before a column
    # in the column - needed for bug in file alignment 2018
    #
    # Also, we can't use * as a comment char, since it is valid
    # in the particle names, as well!

    with filter_file(filename) as file_object:

        nar = pd.read_fwf(
            file_object,
            colspecs=(
                (0, 8),
                (8, 16),
                (16, 24),
                (24, 32),
                (32, 51),
                (51, 60),
                (60, 69),
                (69, 88),
                (88, 97),
                (97, 106),
                (106, 128),
            ),
            header=None,
            names=(
                "ID1",
                "ID2",
                "ID3",
                "ID4",
                "Mass",
                "MassUpper",
                "MassLower",
                "Width",
                "WidthUpper",
                "WidthLower",
                "NameCharge",
            ),
        )

        # Check for duplicates!
        duplicated_ids = (
            nar.duplicated(subset=["ID1"], keep=False) & nar["ID1"].notna()
            | nar.duplicated(subset=["ID2"], keep=False) & nar["ID2"].notna()
            | nar.duplicated(subset=["ID3"], keep=False) & nar["ID3"].notna()
            | nar.duplicated(subset=["ID4"], keep=False) & nar["ID4"].notna()
        )
        if nar[duplicated_ids].shape[0] > 0:
            print("DUPLICATES:\n", nar[duplicated_ids])
        assert (
            nar[duplicated_ids].shape[0] == 0
        ), f"Duplicate entries found in {filename} !"

    ds_list = []
    for i in range(4):
        name = f"ID{i + 1}"
        d = nar[~pd.isna(nar[name])].copy()
        d["ID"] = d[name].astype(int)
        nc = d.NameCharge.str.split(expand=True)
        d["Name"] = nc[0]
        abcd = nc[1].str.split(pat=",", n=4, expand=True)
        d["charge"] = abcd[i]
        d.set_index("ID", inplace=True)
        ds_list.append(d)

    ds = pd.concat(ds_list)
    del ds["NameCharge"], ds["ID1"], ds["ID2"], ds["ID3"], ds["ID4"]
    ds.sort_index(inplace=True)

    # This should be in MeV, not GeV, and absolute value
    for name in ("Mass", "MassUpper", "MassLower", "Width", "WidthUpper", "WidthLower"):
        ds[name] = abs(ds[name] * 1000)

    return ds


def update_from_mcd(
    full_table: pd.DataFrame, update_table: pd.DataFrame
) -> pd.DataFrame:
    """
    Update the full table (aka the PDG extended-style table) with the
    up-to-date information from the PDG .mcd file.

    Example
    -------
    >>> new_table = update_from_mcd('mass_width_2008.fwf', 'mass_width_2022.mcd')    # doctest: +SKIP
    """

    full_table = full_table.copy()
    full_table.update(update_table)
    update_table_neg = update_table.copy()
    update_table_neg.index = -update_table_neg.index
    full_table.update(update_table_neg)

    return full_table


def produce_files(
    particle2008: str | Path,  # pylint: disable=unused-argument
    particle2022: str | Path,
    version: str,
    year: str,
) -> None:
    "This produces listed output files from all input files."

    with data.basepath.joinpath("mass_width_2008.fwf").open() as fwf_f:
        with data.basepath.joinpath("pdgid_to_latexname.csv").open() as csv_f:
            full_table = get_from_pdg_extended(fwf_f, [csv_f])

    # Entries to remove, see comments in file mass_width_2008_ext.fwf:
    # 30221 - the f(0)(1370) since it was renumbered
    # 100223 - the omega(1420) since it was renumbered
    # 5132 and 5232 - the Xi_b baryons got their IDs swapped at some stage
    # 30553 - the Upsilon(2)(1D) was wrongly assigned its ID, and has been renumbered
    full_table.drop([30221, 100223, 5132, 5232], axis=0, inplace=True)

    # No longer write out the particle2008.csv file, which nobody should use
    # with open(particle2008, "w", newline="\n", encoding="utf-8") as f:
    # f.write(version_header(particle2008, version))
    # full_table.to_csv(f, float_format="%.12g")

    with data.basepath.joinpath("mass_width_" + year + ".mcd").open() as mcd_f:
        ext_table = get_from_pdg_mcd(mcd_f)

    with data.basepath.joinpath("mass_width_2008_ext.fwf").open() as fwf_f:
        with data.basepath.joinpath("pdgid_to_latexname.csv").open() as csv_f:
            addons = get_from_pdg_extended(fwf_f, [csv_f])

    # Only keep rows present in the .mcd file specified by year
    full_table = full_table[
        full_table.index.isin(ext_table.index) | full_table.index.isin(-ext_table.index)
    ]

    full_table = pd.concat([full_table, addons])

    # Allow replacement of particles by the ext file
    full_table = full_table[~full_table.index.duplicated(keep="last")]

    sort_particles(full_table)

    new_table = update_from_mcd(full_table, ext_table)

    with open(particle2022, "w", newline="\n", encoding="utf-8") as f:
        f.write(version_header(str(particle2022), version))
        new_table.to_csv(f, float_format="%.12g")


def version_header(filename: str, version_number: str) -> str:
    filename = os.path.basename(filename)
    version = version_number  # version of CSV files
    today_date = date.isoformat(date.today())
    return f"# (c) Scikit-HEP project - Particle package data file - {filename} - version {version} - {today_date}\n"


def main(version: str, year: str) -> None:
    "Regenerate output files - run directly inside the package"
    master_dir = os.path.dirname(FILE_DIR)
    data_dir = os.path.join(master_dir, "data")
    particle2008 = os.path.join(data_dir, "particle2008.csv")
    particlenew = os.path.join(data_dir, "particle" + year + ".csv")

    produce_files(particle2008, particlenew, version, year)


def convert(version: str, output: str, fwf: str, latex: str | None = None) -> None:
    latexes: list[StringOrIO] = [data.basepath / "pdgid_to_latexname.csv"]
    if latex:
        latexes.append(latex)
    table = get_from_pdg_extended(fwf, latexes)

    with open(output, "w", newline="\n", encoding="utf-8") as f:
        f.write(version_header(output, version))
        table.to_csv(f, float_format="%.12g")


def run_regen(args: Any) -> None:
    main(args.version, args.year)


def run_convert(args: Any) -> None:
    convert(args.version, args.output, args.fwf, args.latex)


if __name__ == "__main__":
    from argparse import ArgumentParser, FileType

    parser = ArgumentParser()
    subparsers = parser.add_subparsers(help="Options (pick one)")
    subparsers.required = True
    subparsers.dest = "command"

    parser_regen = subparsers.add_parser(
        "regenerate", help="Regenerate the built in files from the built in names"
    )
    parser_regen.add_argument(
        "year", help="Year of file to read in/produce (2008 is always read/produced)"
    )
    parser_regen.add_argument(
        "version", help="Version of package CSV data files to be written in headers"
    )
    parser_regen.set_defaults(func=run_regen)

    parser_convert = subparsers.add_parser(
        "extended", help="Make a new file from extended inputs"
    )
    parser_convert.add_argument("output", help="Output file name")
    parser_convert.add_argument("fwf", help="Fixed width format extended file")
    parser_convert.add_argument(
        "version", help="Version of package CSV data files to be written in headers"
    )
    parser_convert.add_argument(
        "latex",
        type=FileType("r"),
        help="Optional Latex file with names",
        nargs="?",
        default=None,
    )
    parser_convert.set_defaults(func=run_convert)

    args_ = parser.parse_args()
    args_.func(args_)



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.


from __future__ import annotations

from .corsika import Corsika72PDGIDBiMap
from .evtgen import EvtGen2PDGNameMap, EvtGenName2PDGIDBiMap, PDG2EvtGenNameMap
from .geant import Geant2PDGIDBiMap
from .pythia import Pythia2PDGIDBiMap

__all__ = (
    "EvtGen2PDGNameMap",
    "EvtGenName2PDGIDBiMap",
    "PDG2EvtGenNameMap",
    "Geant2PDGIDBiMap",
    "Pythia2PDGIDBiMap",
    "Corsika72PDGIDBiMap",
)


def __dir__() -> tuple[str, ...]:
    return __all__



# Copyright (c) 2018-2022, Eduardo Rodrigues and Henry Schreiner.
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/scikit-hep/particle for details.

from __future__ import annotations

import pytest

# Requires pandas
pd = pytest.importorskip("pandas")

from pathlib import Path

from particle.particle.convert import get_from_pdg_mcd

DIR = Path(__file__).parent.resolve()


def test_get_from_pdg_mcd():
    with (DIR / "../data/test_PDG_mcd_file_duplicates.mcd").open() as f:
        with pytest.raises(AssertionError):
            get_from_pdg_mcd(f)
